""" HLine module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtWidgets import QFrame


class HLine(QFrame):
    """ HLine class. """

    # Display settings
    VerticalSpacing = 12

    def __init__(self):
        """
        Creates a horizontal line.
        """
        super().__init__()

        self.setFixedHeight(self.VerticalSpacing)
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Plain)
        self.setStyleSheet(f"""
            border-top: 1px solid #cccccc;
            border-bottom: 1px solid #f0f0f0;
            margin-top: {self.VerticalSpacing // 2 - 1}px;
            margin-bottom: {self.VerticalSpacing // 2 - 1}px;
        """)



""" Wire_Widget module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import \
    QVBoxLayout, QHBoxLayout, QPushButton, QSpinBox, QDoubleSpinBox, QComboBox, QLabel, QSizePolicy
from magneticalc.IconLabel import IconLabel
from magneticalc.Groupbox import Groupbox
from magneticalc.HLine import HLine
from magneticalc.ModelAccess import ModelAccess
from magneticalc.Table import Table
from magneticalc.Wire import Wire


class Wire_Widget(Groupbox):
    """ Wire_Widget class. """

    # Display settings
    UnitsLabelWidth = 26

    # Spinbox limits
    RotationalSymmetryCountMin = 1
    RotationalSymmetryCountMax = 99
    RotationalSymmetryRadiusMin = 0
    RotationalSymmetryRadiusMax = 99
    RotationalSymmetryRadiusStep = 0.1
    StretchMin = -99
    StretchMax = 99
    StretchStep = 0.1
    SlicerLimitMinimum = 0.01
    SlicerLimitMaximum = 2.0
    SlicerLimitStep = 0.01
    DcMinimum = -999.0
    DcMaximum = 999.0
    DcStep = 0.1

    def __init__(self, gui):
        """
        Populates the widget.

        @param gui: GUI
        """
        self.gui = gui

        Groupbox.__init__(self, "Wire")

        points_icon_label = IconLabel("mdi.vector-square", "Points")
        table_add_button = QPushButton()
        table_add_button.setIcon(qta.icon("fa.plus"))
        points_icon_label.addWidget(table_add_button)
        table_units_label = QLabel("cm")
        table_units_label.setAlignment(Qt.AlignRight)
        table_units_label.setFixedWidth(self.UnitsLabelWidth)
        points_icon_label.addWidget(table_units_label)
        table_add_button.clicked.connect(self.on_table_row_added)
        self.addWidget(points_icon_label)
        self.table = Table(
            cell_edited_callback=self.on_table_cell_edited,
            row_deleted_callback=self.on_table_row_deleted
        )
        self.table.set_horizontal_header(["X", "Y", "Z"])
        self.addWidget(self.table)

        table_total_layout = QHBoxLayout()
        table_total_label_left = QLabel("Total base points:")
        table_total_label_left.setStyleSheet("""
            color: #555555;
            font-style: italic;
        """)
        self.base_total_label = QLabel("")
        self.base_total_label.setStyleSheet("""
            color: #2a7db0;
        """)
        self.base_total_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        table_total_layout.addWidget(table_total_label_left, alignment=Qt.AlignVCenter)
        table_total_layout.addWidget(self.base_total_label, alignment=Qt.AlignVCenter)
        self.addLayout(table_total_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        stretch_icon_label = IconLabel("mdi.arrow-all", "Stretch")
        stretch_clear_button = QPushButton()
        stretch_clear_button.setIcon(qta.icon("fa.eraser"))
        stretch_clear_button.clicked.connect(self.clear_stretch)
        stretch_icon_label.addWidget(stretch_clear_button)
        stretch_units_label = QLabel("cm")
        stretch_units_label.setAlignment(Qt.AlignRight)
        stretch_units_label.setFixedWidth(self.UnitsLabelWidth)
        stretch_icon_label.addWidget(stretch_units_label)
        self.addWidget(stretch_icon_label)

        self.stretch_spinbox = [None, None, None]
        stretch_label = [None, None, None]
        stretch_layout = QHBoxLayout()
        for i in range(3):
            self.stretch_spinbox[i] = QDoubleSpinBox(self.gui)
            self.stretch_spinbox[i].setMinimum(self.StretchMin)
            self.stretch_spinbox[i].setMaximum(self.StretchMax)
            self.stretch_spinbox[i].setSingleStep(self.StretchStep)
            self.stretch_spinbox[i].setValue(self.gui.config.get_point("wire_stretch")[i])
            self.stretch_spinbox[i].valueChanged.connect(self.set_stretch)
            stretch_label[i] = QLabel(["X", "Y", "Z"][i] + ":")
            stretch_label[i].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
            stretch_layout.addWidget(stretch_label[i], alignment=Qt.AlignVCenter)
            stretch_layout.addWidget(self.stretch_spinbox[i], alignment=Qt.AlignVCenter)
        self.addLayout(stretch_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("mdi.rotate-3d-variant", "Rotational Symmetry"))
        rotational_symmetry_layout = QHBoxLayout()
        rotational_symmetry_layout_left = QVBoxLayout()
        rotational_symmetry_layout_middle = QVBoxLayout()
        rotational_symmetry_layout_right = QVBoxLayout()
        rotational_symmetry_layout.addLayout(rotational_symmetry_layout_left)
        rotational_symmetry_layout.addLayout(rotational_symmetry_layout_middle)
        rotational_symmetry_layout.addLayout(rotational_symmetry_layout_right)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.rotational_symmetry_count_spinbox = QSpinBox()
        self.rotational_symmetry_count_spinbox.setMinimum(self.RotationalSymmetryCountMin)
        self.rotational_symmetry_count_spinbox.setMaximum(self.RotationalSymmetryCountMax)
        self.rotational_symmetry_count_spinbox.setValue(self.gui.config.get_float("rotational_symmetry_count"))
        count_label = QLabel("Count:")
        count_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        rotational_symmetry_layout_left.addWidget(count_label, alignment=Qt.AlignVCenter | Qt.AlignRight)
        rotational_symmetry_layout_middle.addWidget(
            self.rotational_symmetry_count_spinbox,
            alignment=Qt.AlignVCenter
        )
        self.rotational_symmetry_count_spinbox.valueChanged.connect(self.set_rotational_symmetry)
        rotational_symmetric_units_label = QLabel("cm")
        rotational_symmetric_units_label.setAlignment(Qt.AlignRight)
        rotational_symmetric_units_label.setFixedWidth(self.UnitsLabelWidth)
        rotational_symmetry_layout_right.addWidget(rotational_symmetric_units_label, alignment=Qt.AlignVCenter)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.rotational_symmetry_radius_spinbox = QDoubleSpinBox()
        self.rotational_symmetry_radius_spinbox.setMinimum(self.RotationalSymmetryRadiusMin)
        self.rotational_symmetry_radius_spinbox.setMaximum(self.RotationalSymmetryRadiusMax)
        self.rotational_symmetry_radius_spinbox.setSingleStep(self.RotationalSymmetryRadiusStep)
        self.rotational_symmetry_radius_spinbox.setValue(self.gui.config.get_float("rotational_symmetry_radius"))
        radius_label = QLabel("Radius:")
        radius_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        rotational_symmetry_layout_left.addWidget(radius_label, alignment=Qt.AlignVCenter | Qt.AlignRight)
        rotational_symmetry_layout_middle.addWidget(
            self.rotational_symmetry_radius_spinbox,
            alignment=Qt.AlignVCenter
        )
        self.rotational_symmetry_radius_spinbox.valueChanged.connect(self.set_rotational_symmetry)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.rotational_symmetry_axis_combobox = QComboBox()
        for i, axis in enumerate(["X", "Y", "Z"]):
            self.rotational_symmetry_axis_combobox.addItem(axis)
            if i == self.gui.config.get_int("rotational_symmetry_axis"):
                self.rotational_symmetry_axis_combobox.setCurrentIndex(i)
        axis_label = QLabel("Axis:")
        axis_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        rotational_symmetry_layout_left.addWidget(axis_label, alignment=Qt.AlignVCenter | Qt.AlignRight)
        rotational_symmetry_layout_middle.addWidget(
            self.rotational_symmetry_axis_combobox,
            alignment=Qt.AlignVCenter
        )
        self.rotational_symmetry_axis_combobox.currentIndexChanged.connect(self.set_rotational_symmetry)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addLayout(rotational_symmetry_layout)

        rotational_symmetry_total_layout = QHBoxLayout()
        rotational_symmetry_total_label_left = QLabel("Total transformed points:")
        rotational_symmetry_total_label_left.setStyleSheet("""
            color: #555555;
            font-style: italic;
        """)
        self.transformed_total_label = QLabel("")
        self.transformed_total_label.setStyleSheet("""
            color: #2a7db0;
        """)
        self.transformed_total_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        rotational_symmetry_total_layout.addWidget(
            rotational_symmetry_total_label_left,
            alignment=Qt.AlignVCenter
        )
        rotational_symmetry_total_layout.addWidget(self.transformed_total_label, alignment=Qt.AlignVCenter)
        self.addLayout(rotational_symmetry_total_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("mdi.box-cutter", "Slicer Limit"))
        slicer_limit_spinbox = QDoubleSpinBox()
        slicer_limit_spinbox.setMinimum(self.SlicerLimitMinimum)
        slicer_limit_spinbox.setMaximum(self.SlicerLimitMaximum)
        slicer_limit_spinbox.setSingleStep(self.SlicerLimitStep)
        slicer_limit_spinbox.setValue(self.gui.config.get_float("wire_slicer_limit"))
        slicer_limit_spinbox.valueChanged.connect(
            lambda: self.set_wire(slicer_limit=slicer_limit_spinbox.value())
        )
        slicer_limit_units_label = QLabel("cm")
        slicer_limit_units_label.setAlignment(Qt.AlignRight)
        slicer_limit_units_label.setFixedWidth(self.UnitsLabelWidth)
        slicer_limit_layout = QHBoxLayout()
        slicer_limit_layout.addWidget(slicer_limit_spinbox, alignment=Qt.AlignVCenter)
        slicer_limit_layout.addWidget(slicer_limit_units_label, alignment=Qt.AlignVCenter)
        self.addLayout(slicer_limit_layout)

        sliced_total_layout = QHBoxLayout()
        sliced_total_label_left = QLabel("Total sliced points:")
        sliced_total_label_left.setStyleSheet("""
            color: #555555;
            font-style: italic;
        """)
        self.sliced_total_label = QLabel("")
        self.sliced_total_label.setStyleSheet("""
            color: #2a7db0;
        """)
        self.sliced_total_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        sliced_total_layout.addWidget(sliced_total_label_left, alignment=Qt.AlignVCenter)
        sliced_total_layout.addWidget(self.sliced_total_label, alignment=Qt.AlignVCenter)
        self.addLayout(sliced_total_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        dc_spinbox = QDoubleSpinBox()
        self.addWidget(IconLabel("fa.cog", "Wire Current"))
        dc_spinbox.setMinimum(self.DcMinimum)
        dc_spinbox.setMaximum(self.DcMaximum)
        dc_spinbox.setSingleStep(self.DcStep)
        dc_spinbox.setValue(self.gui.config.get_float("wire_dc"))
        dc_spinbox.valueChanged.connect(lambda: self.set_wire(dc=dc_spinbox.value()))
        dc_layout = QHBoxLayout()
        dc_layout.addWidget(dc_spinbox, alignment=Qt.AlignVCenter)
        dc_unit_label = QLabel("A")
        dc_unit_label.setAlignment(Qt.AlignRight)
        dc_unit_label.setFixedWidth(self.UnitsLabelWidth)
        dc_layout.addWidget(dc_unit_label, alignment=Qt.AlignVCenter)
        self.addLayout(dc_layout)

        # Initially load wire from configuration
        self.set_wire(recalculate=False, readjust_sampling_volume=False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_wire(
            self,
            points=None,
            stretch=None,
            rotational_symmetry=None,
            slicer_limit=None,
            dc=None,
            recalculate=True,
            readjust_sampling_volume=True
    ):
        """
        Sets the wire. This will overwrite the currently set wire in the model.
        Any parameter may be left set to None in order to load its default value.

        @param points: Points (List of 3D points)
        @param stretch: XYZ stretch transform factors (3D point)
        @param rotational_symmetry: Dictionary for rotational symmetry transform
        @param slicer_limit: Slicer limit (float)
        @param dc: DC value (float)
        @param recalculate: Enable to trigger final re-calculation (boolean)
        @param readjust_sampling_volume: Enable to readjust sampling volume
        """
        with ModelAccess(self.gui, recalculate):

            points = self.gui.config.set_get_points("wire_points_base", points)
            stretch = self.gui.config.set_get_point("wire_stretch", stretch)
            slicer_limit = self.gui.config.set_get_float("wire_slicer_limit", slicer_limit)
            dc = self.gui.config.set_get_float("wire_dc", dc)

            # ToDo: Implement some kind of "self.gui.config.set_get_dict" to make this shorter
            if rotational_symmetry is None:
                count = self.gui.config.get_int("rotational_symmetry_count")
                radius = self.gui.config.get_float("rotational_symmetry_radius")
                axis = self.gui.config.get_int("rotational_symmetry_axis")
            else:
                count = self.gui.config.set_get_int("rotational_symmetry_count", rotational_symmetry["count"])
                radius = self.gui.config.set_get_float("rotational_symmetry_radius", rotational_symmetry["radius"])
                axis = self.gui.config.set_get_int("rotational_symmetry_axis", rotational_symmetry["axis"])
            rotational_symmetry = {
                "count": count,
                "radius": radius,
                "axis": axis
            }

            self.gui.model.set_wire(
                Wire(
                    points=points,
                    stretch=stretch,
                    rotational_symmetry=rotational_symmetry,
                    slicer_limit=slicer_limit,
                    dc=dc
                )
            )

            self.update_table()

            self.transformed_total_label.setText(str(len(self.gui.model.wire.get_points_transformed())))

            if readjust_sampling_volume:
                self.gui.sidebar_left.sampling_volume_widget.readjust()

    # ------------------------------------------------------------------------------------------------------------------

    def set_stretch(self):
        """
        Handles changes to stretch transform parameters.
        """
        if self.signalsBlocked():
            return

        self.set_wire(stretch=[self.stretch_spinbox[i].value() for i in range(3)])

    def clear_stretch(self):
        """
        Clears the stretch values.
        """
        self.set_wire(stretch=[1.0, 1.0, 1.0])

        self.blockSignals(True)
        for i in range(3):
            self.stretch_spinbox[i].setValue(1.0)
        self.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_rotational_symmetry(self):
        """
        Handles changes to rotational symmetry transform parameters.
        """
        self.set_wire(
            rotational_symmetry={
                "count": self.rotational_symmetry_count_spinbox.value(),
                "radius": self.rotational_symmetry_radius_spinbox.value(),
                "axis": {"X": 0, "Y": 1, "Z": 2}[self.rotational_symmetry_axis_combobox.currentText()]
            }
        )

    # ------------------------------------------------------------------------------------------------------------------

    def update_table(self):
        """
        Populates the wire base points table.
        """
        self.table.clear_rows()

        self.table.set_vertical_header([str(i + 1) for i in range(len(self.gui.model.wire.get_points_base()))])

        points = [[f"{round(col, 2):+0.02f}" for col in row] for row in self.gui.model.wire.get_points_base()]

        self.table.set_contents(points)

        self.base_total_label.setText(str(len(self.gui.model.wire.get_points_base())))

    def on_table_row_added(self):
        """
        Adds a wire base point after a row has been added to the table.
        """
        self.set_wire(points=[np.array([0, 0, 0])] + list(self.gui.model.wire.get_points_base()))

    def on_table_cell_edited(self, item):
        """
        Gets called after a wire point cell has been edited.

        @param item: QTableWidgetItem
        """
        try:
            value = float(item.text().replace(",", "."))
        except ValueError:
            value = 0.0

        points = self.gui.model.wire.get_points_base()
        points[item.row()][item.column()] = value
        self.set_wire(points=points)

    def on_table_row_deleted(self, index):
        """
        Deletes a wire base point after a row has been deleted from the table.

        @param index: Wire base point index
        """
        self.set_wire(points=np.delete(self.gui.model.wire.get_points_base(), index, axis=0))

    # ------------------------------------------------------------------------------------------------------------------

    def update_sliced_total_label(self):
        """
        Updates the total sliced points label.
        Called from calculation thread after the wire became valid.
        """
        if self.gui.model.wire.is_valid():
            self.sliced_total_label.setText(str(len(self.gui.model.wire.get_points_sliced())))
        else:
            self.sliced_total_label.setText("N/A")



""" Metric_Widget module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from si_prefix import si_format
import qtawesome as qta
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QComboBox, QLabel
from magneticalc.IconLabel import IconLabel
from magneticalc.Groupbox import Groupbox
from magneticalc.HLine import HLine
from magneticalc.Metric import Metric
from magneticalc.ModelAccess import ModelAccess


class Metric_Widget(Groupbox):
    """ Metric_Widget class. """

    # Metric limit formatting settings
    LimitsPrecision = 1

    def __init__(self, gui):
        """
        Populates the widget.

        @param gui: GUI
        """
        Groupbox.__init__(self, "Metric")

        self.gui = gui

        # Initially load metric from configuration
        self.set_metric(recalculate=False, update_labels=False)

        self.addWidget(IconLabel("fa.tint", "Color"))
        color_metric_combobox = QComboBox()
        for i, preset in enumerate(Metric.Presets):
            color_metric_combobox.addItem(preset["id"])
            if preset["id"] == self.gui.model.metric.get_color_preset()["id"]:
                color_metric_combobox.setCurrentIndex(i)
        color_metric_combobox.currentIndexChanged.connect(
            lambda: self.set_metric(color_preset=Metric.get_by_id(color_metric_combobox.currentText()))
        )
        self.addWidget(color_metric_combobox)

        self.color_metric_limits_layout = QHBoxLayout()
        # noinspection PyArgumentList
        self.color_metric_limits_widget = QWidget()
        self.color_metric_limits_widget.setLayout(self.color_metric_limits_layout)
        self.color_metric_min_label = QLabel("N/A")
        self.color_metric_max_label = QLabel("N/A")
        color_metric_dots_icon = QLabel()
        color_metric_dots_icon.setPixmap(qta.icon("mdi.dots-horizontal").pixmap(QSize(16, 16)))
        self.color_metric_min_label.setStyleSheet("""background: none;""")
        self.color_metric_max_label.setStyleSheet("""background: none;""")
        color_metric_dots_icon.setStyleSheet("""background: none;""")
        self.color_metric_limits_layout.addWidget(self.color_metric_min_label, alignment=Qt.AlignVCenter)
        self.color_metric_limits_layout.addWidget(color_metric_dots_icon, alignment=Qt.AlignVCenter | Qt.AlignCenter)
        self.color_metric_limits_layout.addWidget(
            self.color_metric_max_label,
            alignment=Qt.AlignVCenter | Qt.AlignRight
        )
        self.addWidget(self.color_metric_limits_widget)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("mdi.blur", "Alpha"))
        alpha_metric_combobox = QComboBox()
        for i, preset in enumerate(Metric.Presets):
            alpha_metric_combobox.addItem(preset["id"])
            if preset["id"] == self.gui.model.metric.get_alpha_preset()["id"]:
                alpha_metric_combobox.setCurrentIndex(i)
        alpha_metric_combobox.currentIndexChanged.connect(
            lambda: self.set_metric(alpha_preset=Metric.get_by_id(alpha_metric_combobox.currentText()))
        )
        self.addWidget(alpha_metric_combobox)

        self.alpha_metric_limits_layout = QHBoxLayout()
        # noinspection PyArgumentList
        self.alpha_metric_limits_widget = QWidget()
        self.alpha_metric_limits_widget.setLayout(self.alpha_metric_limits_layout)
        self.alpha_metric_min_label = QLabel("N/A")
        self.alpha_metric_max_label = QLabel("N/A")
        alpha_metric_dots_icon = QLabel()
        alpha_metric_dots_icon.setPixmap(qta.icon("mdi.dots-horizontal").pixmap(QSize(16, 16)))
        self.alpha_metric_min_label.setStyleSheet("""background: none;""")
        self.alpha_metric_max_label.setStyleSheet("""background: none;""")
        alpha_metric_dots_icon.setStyleSheet("""background: none;""")
        self.alpha_metric_limits_layout.addWidget(self.alpha_metric_min_label, alignment=Qt.AlignVCenter)
        self.alpha_metric_limits_layout.addWidget(alpha_metric_dots_icon, alignment=Qt.AlignVCenter | Qt.AlignCenter)
        self.alpha_metric_limits_layout.addWidget(
            self.alpha_metric_max_label,
            alignment=Qt.AlignVCenter | Qt.AlignRight
        )
        self.addWidget(self.alpha_metric_limits_widget)

    # ------------------------------------------------------------------------------------------------------------------

    def set_metric(self, color_preset=None, alpha_preset=None, recalculate=True, update_labels=True):
        """
        Sets the metric. This will overwrite the currently set metric in the model.
        Any parameter may be left set to None in order to load its default value.

        @param color_preset: Color metric preset (parameters, see Metric module)
        @param alpha_preset: Alpha metric preset (parameters, see Metric module)
        @param recalculate: Enable to trigger final re-calculation (boolean)
        @param update_labels: Enable to update metric labels
        """
        with ModelAccess(self.gui, recalculate):

            # Note: Not using self.gui.config.write_read_str here because we're translating between strings and presets

            if color_preset is None:
                color_preset = Metric.get_by_id(self.gui.config.get_str("color_metric"))
            else:
                self.gui.config.set_str("color_metric", color_preset["id"])

            if alpha_preset is None:
                alpha_preset = Metric.get_by_id(self.gui.config.get_str("alpha_metric"))
            else:
                self.gui.config.set_str("alpha_metric", alpha_preset["id"])

            self.gui.model.set_metric(Metric(color_preset, alpha_preset))

            if update_labels:
                self.update_labels()

    def update_labels(self):
        """
        Updates the metric limit labels.
        """
        if self.gui.model.metric.is_valid():

            limits = self.gui.model.metric.get_limits()

            if self.gui.model.metric.get_color_preset()["is_angle"]:
                color_label_min = "N/A"
                color_label_max = "N/A"

                # ToDo: Show "hsv" colormap gradient instead of nothing at all
                self.color_metric_limits_widget.setStyleSheet("""""")
            else:
                color_label_min = si_format(limits["color_min"], precision=self.LimitsPrecision) + "T"
                color_label_max = si_format(limits["color_max"], precision=self.LimitsPrecision) + "T"
                self.color_metric_limits_widget.setStyleSheet("""
                    background: qlineargradient(x1:0 y1:0, x2:1 y2:0, stop:0 #00fffe, stop:1 #ff22f9);
                """)

            if self.gui.model.metric.get_alpha_preset()["is_angle"]:
                alpha_label_min = "N/A"
                alpha_label_max = "N/A"

                # ToDo: Show "hsv" colormap gradient instead of nothing at all
                self.alpha_metric_limits_widget.setStyleSheet("""""")
            else:
                alpha_label_min = si_format(limits["alpha_min"], precision=self.LimitsPrecision) + "T"
                alpha_label_max = si_format(limits["alpha_max"], precision=self.LimitsPrecision) + "T"
                self.alpha_metric_limits_widget.setStyleSheet("""
                    background: qlineargradient(x1:0 y1:0, x2:1 y2:0, stop:0 #00fffe, stop:1 #ff22f9);
                """)

            self.color_metric_min_label.setText(color_label_min)
            self.color_metric_max_label.setText(color_label_max)
            self.alpha_metric_min_label.setText(alpha_label_min)
            self.alpha_metric_max_label.setText(alpha_label_max)

        else:
            self.invalidate_labels()

    def invalidate_labels(self):
        """
        Invalidates the metric limit labels.
        """
        self.color_metric_min_label.setText("N/A")
        self.color_metric_max_label.setText("N/A")
        self.alpha_metric_min_label.setText("N/A")
        self.alpha_metric_max_label.setText("N/A")
        self.color_metric_limits_widget.setStyleSheet("""""")
        self.alpha_metric_limits_widget.setStyleSheet("""""")



""" Assert_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QTextBrowser, QPushButton
from magneticalc.Debug import Debug
from magneticalc.Version import Version
from urllib.parse import urlencode, quote_plus


class Assert_Dialog:
    """ Assert_Dialog class. """

    # Window dimensions
    Width = 600

    def __init__(self, assertion, message):
        """
        Shows a user dialog if an assertion failed. Intended for beta-testing.
        This allows the user to either quit or resume (possibly resulting in unstable behaviour).
        Furthermore, it provides a link for filing an issue on GitHub (partially pre-filled).

        @param assertion: Boolean
        @param message: Error message
        """
        self.assertion = assertion
        self.message = message

        if assertion:
            return

        Debug(self, f": Failed: {message}", color=(255, 0, 0), force=True)

        # noinspection PyArgumentList
        self.dialog = QDialog()

        self.dialog.setWindowTitle("Assertion failed")

        layout = QVBoxLayout()
        self.dialog.setLayout(layout)

        # Generate Github issue URL
        issue_url = \
            f"https://github.com/shredEngineer/MagnetiCalc/issues/new?" + \
            urlencode(
                {
                    "title": f"Assertion failed: {self.message}",
                    "labels": "bug",
                    "body":
                        f"**I discovered a bug in {Version.String}:**\n"
                        "```\n"
                        f"{self.message}\n"
                        "```\n\n"
                        f"**Steps to reproduce the problem:**\n\n"
                        "[_Please make sure to fill this in!_]"
                },
                quote_via=quote_plus
            )

        # HTML content
        html = f"""
            <span style="color: #2a7db0;"><b>Sorry for the inconvenience!</b></span><br>
            <br>
            You seem to have discovered a bug in MagnetiCalc.<br>
            If this error persists, please
            <a href="{issue_url}">file an issue on GitHub</a>!<br>
            <br>
            <b>Error message:</b><br>
            <pre>{self.message}</pre>
            """

        text_browser = QTextBrowser()
        text_browser.setMinimumWidth(self.Width)
        text_browser.setStyleSheet("""
            background: palette(window);
            border: none;
            line-height: 20px;
        """)
        text_browser.setOpenExternalLinks(True)
        text_browser.insertHtml(html)
        text_browser.setFocusPolicy(Qt.NoFocus)
        cursor = text_browser.textCursor()
        cursor.setPosition(0)
        text_browser.setTextCursor(cursor)
        layout.addWidget(text_browser, alignment=Qt.AlignTop)

        button_box = QHBoxLayout()

        ok_button = QPushButton(qta.icon("fa.times-circle"), "Abort application")
        ok_button.clicked.connect(self.reject)
        button_box.addWidget(ok_button, alignment=Qt.AlignBottom)

        donate_button = QPushButton(qta.icon("fa.play-circle"), "Resume (possibly unstable)")
        donate_button.clicked.connect(self.accept)
        button_box.addWidget(donate_button, alignment=Qt.AlignBottom)

        layout.addLayout(button_box)

        self.dialog.exec()

    def reject(self):
        """
        User chose to abort.
        """
        self.dialog.reject()

        sys.exit()

    def accept(self):
        """
        User chose to resume.
        """
        self.dialog.accept()



""" SidebarRight module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QScrollArea, QVBoxLayout, QSizePolicy
from magneticalc.Display_Widget import Display_Widget
from magneticalc.Metric_Widget import Metric_Widget
from magneticalc.Perspective_Widget import Perspective_Widget


class SidebarRight(QScrollArea):
    """ SidebarRight class. """

    # Display settings
    MaximumWidth = 380
    VerticalSpacing = 12

    def __init__(self, gui):
        """
        Populates the right sidebar.

        @param gui: GUI
        """
        QScrollArea.__init__(self)

        self.gui = gui

        layout = QVBoxLayout()
        # noinspection PyArgumentList
        widget = QWidget()
        widget.setLayout(layout)
        self.setWidget(widget)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setWidgetResizable(True)
        self.setMaximumWidth(self.MaximumWidth)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)

        # --------------------------------------------------------------------------------------------------------------

        self.metric_widget = Metric_Widget(self.gui)
        layout.addWidget(self.metric_widget, alignment=Qt.AlignTop)

        layout.addSpacing(self.VerticalSpacing)

        # --------------------------------------------------------------------------------------------------------------

        self.perspective_widget = Perspective_Widget(self.gui)
        layout.addWidget(self.perspective_widget, alignment=Qt.AlignTop)

        layout.addSpacing(self.VerticalSpacing)

        # --------------------------------------------------------------------------------------------------------------

        self.display_widget = Display_Widget(self.gui)
        layout.addWidget(self.display_widget, alignment=Qt.AlignTop)

        layout.addStretch()



""" Metric module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
import matplotlib.cm as cm
from matplotlib.colors import Normalize, LogNorm
from PyQt5.QtCore import QThread
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Debug import Debug


class Metric:
    """
    Provides different metrics, mapping some field vector properties to some color and alpha range.
    For a list of color maps, see: https://matplotlib.org/3.1.1/gallery/color/colormap_reference.html
    """

    # Length scale
    LengthScale = 1e-2  # m  (== 1 cm)

    # Cutoff for logarithmic scaling
    LogNormMinimum = 1e-12

    # Metric value: Magnitude in XYZ-space (linear)
    Magnitude = {
        "id": "Magnitude",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(point),
        "log": False,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric value: Magnitude in XY-plane (linear)
    MagnitudeXY = {
        "id": "Magnitude XY",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(np.array([point[0], point[1]])),
        "log": False,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric value: Magnitude in XZ-plane (linear)
    MagnitudeXZ = {
        "id": "Magnitude XZ",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(np.array([point[0], point[2]])),
        "log": False,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric value: Magnitude in YZ-plane (linear)
    MagnitudeYZ = {
        "id": "Magnitude YZ",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(np.array([point[1], point[2]])),
        "log": False,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric value: Magnitude in XYZ-space (logarithmic)
    LogMagnitude = {
        "id": "Log Magnitude",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(point),
        "log": True,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric value: Magnitude in XY-plane (logarithmic)
    LogMagnitudeXY = {
        "id": "Log Magnitude XY",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(np.array([point[0], point[1]])),
        "log": True,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric value: Magnitude in XZ-plane (logarithmic)
    LogMagnitudeXZ = {
        "id": "Log Magnitude XZ",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(np.array([point[0], point[2]])),
        "log": True,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric preset: Magnitude in YZ-plane (logarithmic)
    LogMagnitudeYZ = {
        "id": "Log Magnitude YZ",
        "func": lambda point: Metric.LengthScale * np.linalg.norm(np.array([point[1], point[2]])),
        "log": True,
        "is_angle": False,
        "colormap": cm.cool  # divergent
    }

    # Metric preset: Angle in XY-plane
    AngleXY = {
        "id": "Angle XY",
        "func": lambda point: (np.arctan2(point[0], point[1]) + np.pi) / np.pi / 2,
        "log": False,
        "is_angle": True,
        "colormap": cm.hsv  # cyclic
    }

    # Metric preset: Angle in XZ-plane
    AngleXZ = {
        "id": "Angle XZ",
        "func": lambda point: (np.arctan2(point[0], point[2]) + np.pi) / np.pi / 2,
        "log": False,
        "is_angle": True,
        "colormap": cm.hsv  # cyclic
    }

    # Metric preset: Angle in YZ-plane
    AngleYZ = {
        "id": "Angle YZ",
        "func": lambda point: (np.arctan2(point[1], point[2]) + np.pi) / np.pi / 2,
        "log": False,
        "is_angle": True,
        "colormap": cm.hsv  # cyclic
    }

    # List of all above presets
    Presets = [
        Magnitude,
        MagnitudeXY,
        MagnitudeXZ,
        MagnitudeYZ,
        LogMagnitude,
        LogMagnitudeXY,
        LogMagnitudeXZ,
        LogMagnitudeYZ,
        AngleXY,
        AngleXZ,
        AngleYZ
    ]

    @staticmethod
    def get_by_id(_id_):
        """
        Selects a preset by name.

        @param _id_: Preset ID string
        @return: Preset parameters (or None if ID not found)
        """
        for preset in Metric.Presets:
            if _id_ == preset["id"]:
                return preset
        return None

    def __init__(self, color_preset, alpha_preset):
        """
        This class holds a pair of metric presets.
        Using these metric presets, colors (including alpha channel) and field limits are calculated.

        @param color_preset: Color metric preset (dictionary)
        @param alpha_preset: Alpha metric preset (dictionary)
        """
        Debug(self, ": Init")

        self._color_preset = color_preset
        self._alpha_preset = alpha_preset

        self._colors = None
        self._limits = None

    def is_valid(self):
        """
        Indicates valid data for display.

        @return: True if data is valid for display, False otherwise
        """
        return \
            self._colors is not None and \
            self._limits is not None

    def invalidate(self):
        """
        Resets data, hiding from display.
        """
        Debug(self, ".invalidate()", color=(128, 0, 0))

        self._colors = None
        self._limits = None

    def get_color_preset(self):
        """
        Returns color metric preset.

        @return: Color metric preset (dictionary)
        """
        return self._color_preset

    def get_alpha_preset(self):
        """
        Returns alpha metric preset.

        @return: Alpha metric preset (dictionary)
        """
        return self._alpha_preset

    def get_colors(self):
        """
        Returns calculated colors.

        @return: List of color values (4-tuples)
        """
        return self._colors

    def get_limits(self):
        """
        Returns calculated limits.

        @return: Dictionary
        """
        return self._limits

    def recalculate(self, field_vectors, progress_callback):
        """
        Recalculate color and alpha values for field.

        @param field_vectors: Field vectors
        @param progress_callback: Progress callback
        @return: True if successful, False if interrupted
        """
        Debug(self, ".recalculate()", color=(0, 128, 0))

        # Calculate color and alpha metric values
        color_values = np.zeros(len(field_vectors))
        alpha_values = np.zeros(len(field_vectors))
        for i in range(len(field_vectors)):
            color_values[i] = self._color_preset["func"](field_vectors[i])
            alpha_values[i] = self._alpha_preset["func"](field_vectors[i])

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                progress_callback(10 * (i + 1) / len(field_vectors))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): Interruption requested, exiting now", color=(0, 0, 255))
                    return False

        # Select color range
        if self._color_preset["is_angle"]:
            color_min, color_max = 0, 1
        else:
            color_min, color_max = min(color_values), max(color_values)

        # Select alpha range
        if self._alpha_preset["is_angle"]:
            alpha_min, alpha_max = 0, 1
        else:
            alpha_min, alpha_max = min(alpha_values), max(alpha_values)

        # Select color normalizer
        if self._color_preset["log"]:
            Assert_Dialog(not self._color_preset["is_angle"], "Logarithmic angles don't make any sense")

            # Adjust range for logarithm (out-of-range values will be clipped in the loop below)
            color_min_ = max(color_min, Metric.LogNormMinimum)   # avoiding "ValueError: minvalue must be positive"
            color_max_ = max(color_min_, color_max)              # avoiding "ValueError: minvalue must be <= maxvalue"

            color_normalize = LogNorm(vmin=color_min_, vmax=color_max_)
        else:
            color_min_ = color_min
            color_max_ = color_max
            color_normalize = Normalize(vmin=color_min_, vmax=color_max_)

        # Select alpha normalizer
        if self._alpha_preset["log"]:
            Assert_Dialog(not self._alpha_preset["is_angle"], "Logarithmic angles don't make any sense")

            # Adjust range for logarithm (out-of-range values will be clipped in the loop below)
            alpha_min_ = max(alpha_min, Metric.LogNormMinimum)   # avoiding "ValueError: minvalue must be positive"
            alpha_max_ = max(alpha_min_, alpha_max)              # avoiding "ValueError: minvalue must be <= maxvalue"

            alpha_normalize = LogNorm(vmin=alpha_min_, vmax=alpha_max_)
        else:
            alpha_min_ = alpha_min
            alpha_max_ = alpha_max
            alpha_normalize = Normalize(vmin=alpha_min_, vmax=alpha_max_)

        # (1) Calculate color and alpha normalization
        # (2) Calculate colormap and assemble final color/alpha values
        #     Note: Of course, the alpha metric colormap is ignored
        colors = np.zeros([len(field_vectors), 4])
        for i in range(len(field_vectors)):

            # Calculate normalized color and alpha values
            # Note: If using logarithmic scaling, out-of-range values (still exceeding [0...1] here) may be clipped
            color_value_normalized = color_normalize(color_values[i], clip=True)
            alpha_value_normalized = alpha_normalize(alpha_values[i], clip=True)

            colors[i] = self._color_preset["colormap"](color_value_normalized)
            colors[i][3] = alpha_value_normalized

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                progress_callback(10 + 90 * (i + 1) / len(field_vectors))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): Interruption requested, exiting now", color=(0, 0, 255))
                    return False

        self._colors = colors

        self._limits = {
            "color_min": color_min,
            "color_max": color_max,
            "alpha_min": alpha_min,
            "alpha_max": alpha_max
        }

        return True



""" GUI module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import time
import atexit
import datetime
import qtawesome as qta
from PyQt5.QtCore import Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import QApplication, QMainWindow, QSplitter, QFileDialog
from magneticalc.CalculationThread import CalculationThread
from magneticalc.Debug import Debug
from magneticalc.Menu import Menu
from magneticalc.Model import Model
from magneticalc.SidebarLeft import SidebarLeft
from magneticalc.SidebarRight import SidebarRight
from magneticalc.Statusbar import Statusbar
from magneticalc.Version import Version
from magneticalc.VispyCanvas import VispyCanvas


class GUI(QMainWindow):
    """ GUI class. """

    # These signals are fired from the from calculation thread
    calculation_status = pyqtSignal(str)
    calculation_exited = pyqtSignal(bool)

    def __init__(self, config):
        """
        Initializes the GUI.

        @param config: Config
        """
        QMainWindow.__init__(self, flags=Qt.Window)

        Debug(self, ": Init")

        self.config = config

        self.set_window()

        # The calculation thread is started once initially; after that, recalculation is triggered through ModelAccess
        self.calculation_thread = None  # Will be initialized by self.recalculate() but is needed here for ModelAccess
        self.calculation_start_time = None

        # Register exit handler (used by Assert_Dialog to exit gracefully)
        atexit.register(self.quit)

        # Create the model first, as the following objects will access it (each widget acts as view *and* controller)
        self.model = Model(
            # This callback is needed to clear the metric labels after the metric became invalid
            on_metric_invalidated=self.on_metric_invalidated
        )

        # Create the left and right sidebar
        # Note: These create the wire, sampling volume and metric widgets, each populating the model from configuration
        self.sidebar_left = SidebarLeft(self)
        self.sidebar_right = SidebarRight(self)

        # Create the VisPy canvas (our 3D scene) and statusbar
        self.vispy_canvas = VispyCanvas(self)
        self.statusbar = Statusbar(self)

        # Insert left sidebar, VisPy canvas and right sidebar into main layout.
        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.addWidget(self.sidebar_left)
        self.splitter.addWidget(self.vispy_canvas.native)
        self.splitter.addWidget(self.sidebar_right)
        self.setCentralWidget(self.splitter)
        self.splitter.setHandleWidth(8)

        # Insert the menu
        self.menu = Menu(self)

        # Connect the calculation thread communication signals
        self.calculation_status.connect(lambda text: self.statusbar.text(text))
        self.calculation_exited.connect(lambda success: self.on_calculation_exited(success))

        # Kick off the field calculation
        self.recalculate()

    # ------------------------------------------------------------------------------------------------------------------

    def redraw(self):
        """
        Re-draws the scene.
        """
        if self.calculation_thread is not None:
            if self.calculation_thread.isRunning():
                Debug(self, ".redraw(): Skipped because calculation is in progress", color=(0, 0, 255))
                return
            else:
                Debug(self, ".redraw(): WARNING: Setting calculation thread to None", color=(255, 0, 0))
                self.calculation_thread = None

        self.vispy_canvas.redraw()

    # ------------------------------------------------------------------------------------------------------------------

    def recalculate(self):
        """
        Re-calculates the model.
        """
        Debug(self, ".recalculate()", color=(0, 128, 0))

        if self.calculation_thread is not None:
            Debug(self, ".recalculate(): WARNING: Killing orphaned calculation thread", color=(255, 0, 0))
            self.interrupt_calculation()

        self.redraw()
        self.statusbar.arm()

        # Create a new calculation thread and kick it off
        self.calculation_thread = CalculationThread(self)
        self.calculation_start_time = time.monotonic()
        self.calculation_thread.start()

    def on_calculation_exited(self, success):
        """
        This is called after calculation thread has exited.

        @param success: Reflects successful calculation
        """
        calculation_time = time.monotonic() - self.calculation_start_time

        if self.calculation_thread is not None:
            if self.calculation_thread.isRunning():
                # Skipping because another thread is now running
                # Note: This happens all the time when calculation is interrupted and restarted through ModelAccess;
                #       we see this because there is no reliable way to revoke the delayed "calculation_exited" signal
                #       after another thread has already been started
                return
            else:
                # This happens when calculation finished and there
                self.calculation_thread = None
                Debug(self, f".on_calculation_exited(): Success (took {calculation_time:.2f} s)", color=(0, 128, 0))
        else:
            Debug(self, f".on_calculation_exited(): Interrupted after {calculation_time:.2f} s", color=(0, 0, 255))

        # Note: For some reason, most of the time we need an additional ("final-final") re-draw here
        self.redraw()

        self.statusbar.disarm(success)

    def interrupt_calculation(self):
        """
        Kills any running calculation.
        """
        if self.calculation_thread is None:
            Debug(self, ".interrupt_calculation: WARNING: No calculation thread to interrupt", color=(255, 0, 0))
            return

        if self.calculation_thread.isRunning():
            Debug(self, ".interrupt_calculation(): Requesting interruption", color=(0, 0, 255))
            self.calculation_thread.requestInterruption()

            if self.calculation_thread.wait(1000):
                Debug(self, ".interrupt_calculation(): Exited gracefully", color=(0, 0, 255))
            else:
                Debug(self, ".interrupt_calculation(): WARNING: Terminating ungracefully", color=(255, 0, 0))
                self.calculation_thread.terminate()
                self.calculation_thread.wait()
        else:
            Debug(self, ".interrupt_calculation: WARNING: Calculation thread should be running", color=(255, 0, 0))

        self.calculation_thread = None

    # ------------------------------------------------------------------------------------------------------------------

    def on_metric_invalidated(self):
        """
        Needed to clear the metric labels after the metric became invalid.
        """
        self.sidebar_right.metric_widget.invalidate_labels()

    # ------------------------------------------------------------------------------------------------------------------

    def set_window(self):
        """
        Sets the basic window properties.
        """

        # Set window title, icon and dimensions
        self.setWindowTitle(Version.String)
        self.setWindowIcon(qta.icon("ei.magnet"))
        self.resize(self.config.get_int("window_width"), self.config.get_int("window_height"))

        # Center window
        fg = self.frameGeometry()
        # noinspection PyArgumentList
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        # noinspection PyArgumentList
        fg.moveCenter(QApplication.desktop().screenGeometry(screen).center())
        self.move(fg.topLeft())

    def quit(self):
        """
        Quits the application.
        """
        if self.calculation_thread != QThread.currentThread():
            Debug(self, ".quit()")
            if self.calculation_thread is not None:
                self.interrupt_calculation()
        else:
            Debug(self, ".quit(): Called from calculation thread (assertion failed)")

        self.config.set_float("azimuth", self.vispy_canvas.view_main.camera.azimuth)
        self.config.set_float("elevation", self.vispy_canvas.view_main.camera.elevation)
        self.config.set_float("scale_factor", self.vispy_canvas.view_main.camera.scale_factor)

        self.config.close()

        print()
        print("Goodbye!")

        # Unregister exit handler (used by Assert_Dialog to exit gracefully)
        atexit.unregister(self.quit)

    def closeEvent(self, _event):
        """
        Handle close event.

        @param _event: Close event
        """
        self.quit()

    def file_save(self):
        """
        Saves the currently displayed scene to PNG file.
        """

        # noinspection PyCallByClass,PyTypeChecker
        filename, _chosen_extension = QFileDialog.getSaveFileName(
            self,
            "Save Image",
            datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S_MagnetiCalc"),
            "Portable Network Graphics (*.png)",
            options=QFileDialog.DontUseNativeDialog
        )

        if filename != "":
            _file_name, file_extension = os.path.splitext(filename)

            if file_extension.lower() != ".png":
                filename += ".png"

            self.vispy_canvas.save_image(filename)



""" Debug module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from inspect import isclass, stack
from colorit import color_front, bold


class Debug:
    """ Debug class. """

    # Enable debug output from specific classes only
    Whitelist = [
        "About_Dialog",
        "Assert_Dialog",
        "BiotSavart",
        "CalculationThread",
        # "Config",
        "Display_Widget",
        "Field",
        "Groupbox",
        "GUI",
        "HLine",
        "IconLabel",
        "Menu",
        "Metric",
        "Metric_Widget",
        "Model",
        "ModelAccess",
        "Perspective_Widget",
        "SamplingVolume",
        "SamplingVolume_Widget",
        "SidebarLeft",
        "SidebarRight",
        "SliderFloat",
        "Statusbar",
        "Table",
        "Usage_Dialog",
        "Version",
        "VispyCanvas",
        "Wire",
        "Wire_Widget"
    ]

    def __init__(self, obj, text, color=None, force=False):
        """
        Displays a colorful debug message and the current call hierarchy.

        @param obj: Class instance
        @param text: Debug message
        @param color: Color (may be None)
        @param force: Enable to override whitelist
        """
        if isclass(obj):
            # Called from within class method, i.e. Debug(cls, ...)
            name = obj.__name__
        else:
            # Called from within instance method, i.e. Debug(self, ...)
            name = type(obj).__name__

        if not force:
            # Skip non-whitelisted class names
            if name not in self.Whitelist:
                return

        # A class may specify its own default color
        if color is None:
            if hasattr(obj, "DebugColor"):
                color = obj.DebugColor
            else:
                color = (0, 0, 0)

        # Format call hierarchy
        hierarchy = ""
        for x in reversed(stack()[2:]):
            func_str = str(x.function)
            if func_str in ["<module>", "_run_module_as_main", "_run_code"]:
                # Called from within the main module (don't display these elements)
                continue
            if func_str == "run":
                # Called from within another thread (it is the calculation thread)
                func_str = "\tCalculationThread"
            hierarchy += func_str + "/"

        if color == (0, 0, 0):
            # Allow a terminal use its own foreground color
            color_text = bold(name) + text
        else:
            color_text = color_front(bold(name), *color) + color_front(text, *color)

        print(color_front(hierarchy, 128, 128, 128) + color_text + "\n", end="")



""" Menu module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import qtawesome as qta
from functools import partial
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMenu, QAction
from magneticalc.About_Dialog import About_Dialog
from magneticalc.Usage_Dialog import Usage_Dialog
from magneticalc.Wire import Wire


class Menu:
    """ Menu class. """

    def __init__(self, gui):
        """
        Creates the menu bar.

        @param gui: GUI
        """
        self.gui = gui

        # List of checkboxes that are bound to configuration
        self.config_bound_checkboxes = {}

        # File menu
        file_menu = QMenu("&File", self.gui)
        file_menu.addSeparator()
        file_menu.addAction(qta.icon("fa.picture-o"), "&Save Image...", self.gui.file_save, Qt.CTRL + Qt.Key_S)
        file_menu.addSeparator()
        file_menu.addAction(qta.icon("fa.window-close"), "&Quit", self.gui.close, Qt.CTRL + Qt.Key_Q)
        self.gui.menuBar().addMenu(file_menu)

        # Wire menu
        wire_menu = QMenu("&Load Wire Preset", self.gui)
        for preset in Wire.Presets:
            action = QAction(preset["id"], wire_menu)
            action.setIcon(qta.icon("mdi.vector-square"))
            action.triggered.connect(
                partial(self.gui.sidebar_left.wire_widget.set_wire, points=preset["points"])
            )
            wire_menu.addAction(action)
        self.gui.menuBar().addMenu(wire_menu)

        # View menu
        view_menu = QMenu("&View", self.gui)
        self.add_config_bound_checkbox("Show Wire Segments", "show_wire_segments", view_menu, self.gui.redraw)
        self.add_config_bound_checkbox("Show Wire Points", "show_wire_points", view_menu, self.gui.redraw)
        view_menu.addSeparator()
        self.add_config_bound_checkbox("Use Field Colors for Labels", "field_colors_labels", view_menu, self.gui.redraw)
        view_menu.addSeparator()
        self.add_config_bound_checkbox("Show Coordinate System", "show_coordinate_system", view_menu, self.gui.redraw)
        self.add_config_bound_checkbox("Show Perspective Info", "show_perspective_info", view_menu, self.gui.redraw)
        view_menu.addSeparator()
        self.add_config_bound_checkbox("Dark Background", "dark_background", view_menu, self.gui.redraw)
        self.gui.menuBar().addMenu(view_menu)

        # Help menu
        help_menu = QMenu("&Help", self.gui)
        help_menu.addAction(qta.icon("fa.info"), "&Usage", lambda: Usage_Dialog().show(), Qt.Key_F1)
        help_menu.addSeparator()
        help_menu.addAction(qta.icon("fa.coffee"), "&About", lambda: About_Dialog().show())
        self.gui.menuBar().addMenu(help_menu)

    def add_config_bound_checkbox(self, label, key, menu, callback):
        """
        Creates a checkbox inside some menu. Checkbox state is bound to configuration.

        @param label: Checkbox label
        @param key: Configuration key
        @param menu: Menu
        @param callback:
        """
        checkbox = QAction(label, menu)
        checkbox.setCheckable(True)
        checkbox.triggered.connect(partial(self.config_bound_checkbox_changed, key))
        self.config_bound_checkboxes[key] = {"checkbox": checkbox, "callback_final": callback}
        checkbox.setChecked(self.gui.config.get_bool(key))
        menu.addAction(checkbox)

    def config_bound_checkbox_changed(self, key):
        """
        Handles change of checkbox state.

        @param key: Configuration key
        """
        self.gui.config.set_bool(key, self.config_bound_checkboxes[key]["checkbox"].isChecked())
        self.config_bound_checkboxes[key]["callback_final"]()



""" SamplingVolume_Widget module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QPushButton, QSpinBox, QSizePolicy
from magneticalc.Debug import Debug
from magneticalc.IconLabel import IconLabel
from magneticalc.Groupbox import Groupbox
from magneticalc.HLine import HLine
from magneticalc.ModelAccess import ModelAccess
from magneticalc.SamplingVolume import SamplingVolume


class SamplingVolume_Widget(Groupbox):
    """ SamplingVolume_Widget class. """

    # Display settings
    UnitsLabelWidth = 26

    # Spinbox limits
    PaddingMin = -99
    PaddingMax = 99
    ResolutionMinimum = 1
    ResolutionMaximum = 50

    def __init__(self, gui):
        """
        Populates the widget.

        @param gui: GUI
        """
        Groupbox.__init__(self, "Sampling Volume")

        self.gui = gui

        padding_icon_label = IconLabel("mdi.arrow-expand-all", "Padding")
        padding_clear_button = QPushButton()
        padding_clear_button.setIcon(qta.icon("fa.eraser"))
        padding_clear_button.clicked.connect(self.clear_padding)
        padding_icon_label.addWidget(padding_clear_button)
        padding_units_label = QLabel("cm")
        padding_units_label.setAlignment(Qt.AlignRight)
        padding_units_label.setFixedWidth(self.UnitsLabelWidth)
        padding_icon_label.addWidget(padding_units_label)
        self.addWidget(padding_icon_label)

        padding_layout = QHBoxLayout()
        padding_label = [None, None, None]
        self.padding_spinbox = [None, None, None]
        for i in range(3):
            self.padding_spinbox[i] = QSpinBox(self.gui)
            self.padding_spinbox[i].setMinimum(self.PaddingMin)
            self.padding_spinbox[i].setMaximum(self.PaddingMax)
            self.padding_spinbox[i].setValue(self.gui.config.get_point("sampling_volume_padding")[i])
            self.padding_spinbox[i].valueChanged.connect(self.update_padding)
            padding_label[i] = QLabel(["X", "Y", "Z"][i] + ":")
            padding_label[i].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
            padding_layout.addWidget(padding_label[i], alignment=Qt.AlignVCenter)
            padding_layout.addWidget(self.padding_spinbox[i], alignment=Qt.AlignVCenter)
        self.addLayout(padding_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("fa.th", "Resolution"))
        resolution_spinbox = QSpinBox(self.gui)
        resolution_spinbox.setMinimum(self.ResolutionMinimum)
        resolution_spinbox.setMaximum(self.ResolutionMaximum)
        resolution_spinbox.setValue(self.gui.config.get_int("sampling_volume_resolution"))
        resolution_spinbox.valueChanged.connect(
            lambda: self.set_sampling_volume(resolution=resolution_spinbox.value())
        )
        resolution_layout = QHBoxLayout()
        resolution_layout.addWidget(resolution_spinbox, alignment=Qt.AlignVCenter)
        points_units_label = QLabel("Points / cm")
        points_units_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        resolution_layout.addWidget(points_units_label, alignment=Qt.AlignVCenter)
        self.addLayout(resolution_layout)

        total_layout = QHBoxLayout()
        total_label_left = QLabel("Total constrained points:")
        total_label_left.setStyleSheet("""
            color: #555555;
            font-style: italic;
        """)
        self.total_label = QLabel("")
        self.total_label.setStyleSheet("""
            color: #2a7db0;
        """)
        self.total_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        total_layout.addWidget(total_label_left, alignment=Qt.AlignVCenter)
        total_layout.addWidget(self.total_label, alignment=Qt.AlignVCenter)
        self.addLayout(total_layout)

        # ToDo: Add a table to add, edit and remove sampling volume constraints

        # Initially load sampling volume from configuration
        self.set_sampling_volume(recalculate=False)

    # ------------------------------------------------------------------------------------------------------------------

    def update_padding(self):
        """
        Updates padding.
        """
        if self.signalsBlocked():
            return

        self.set_sampling_volume(padding=[self.padding_spinbox[i].value() for i in range(3)])

    def clear_padding(self):
        """
        Clears the padding values.
        """
        self.set_sampling_volume(padding=[0, 0, 0])

        self.blockSignals(True)
        for i in range(3):
            self.padding_spinbox[i].setValue(0)
        self.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_sampling_volume(
            self,
            padding=None,
            resolution=None,
            recalculate=True
    ):
        """
        Sets the sampling volume. This will overwrite the currently set sampling volume in the model.
        Parameters may be left set to None in order to load their default value.

        @param resolution: Sampling volume _resolution
        @param padding: Padding (3D point)
        @param recalculate: Enable to trigger final re-calculation (boolean)
        """
        with ModelAccess(self.gui, recalculate):

            resolution = self.gui.config.set_get_int("sampling_volume_resolution", resolution)

            self.gui.model.set_sampling_volume(SamplingVolume(resolution))

            self.readjust(padding)

            # ToDo: Add support for sampling volume constraints, e.g.:
            # self.gui.model.sampling_volume.add_constraint(lambda p: -1.5 <= p["z"] <= 1.5)
            # self.gui.model.sampling_volume.add_constraint(lambda p: 2 <= p["radius_xy"] <= 4.5)
            # self.gui.model.sampling_volume.add_constraint(lambda p: p["y"] == 0)
            # self.gui.model.sampling_volume.add_constraint(lambda p: p["radius"] <= 3)

    # ------------------------------------------------------------------------------------------------------------------

    def readjust(self, padding=None):
        """
        Readjusts the sampling volume to the currently set wire bounds and padding.
        The parameter may be left set to None in order to load its default value.
        This also readjusts the minimum padding bounds in case the wire bounds have shrunk too far.

        @param padding: Padding (3D point), may be None to be read from configuration
        """
        Debug(self, ".readjust()")

        self.gui.model.sampling_volume.set_bounds_nearest(*self.gui.model.wire.get_bounds())

        bounds_min, bounds_max = self.gui.model.sampling_volume.get_bounds()

        # Adjust padding values if they are now outside the new minimum padding bounds
        padding = self.gui.config.set_get_point("sampling_volume_padding", padding)
        padding_adjusted = False
        for i in range(3):
            if bounds_min[i] > self.padding_spinbox[i].value():
                padding[i] = bounds_min[i]
                padding_adjusted = True
        if padding_adjusted:
            Debug(self, ".readjust(): Adjusted padding values to new minimum padding bounds")
            self.gui.config.set_point("sampling_volume_padding", padding)

        self.blockSignals(True)
        for i in range(3):
            self.padding_spinbox[i].setMinimum(bounds_min[i])
        self.blockSignals(False)

        self.gui.model.sampling_volume.set_padding(*padding)

    # ------------------------------------------------------------------------------------------------------------------

    def update_total_label(self):
        """
        Updates the total sampling volume points label.
        Called from calculation thread after the sampling volume became valid.
        """
        if self.gui.model.sampling_volume.is_valid():
            self.total_label.setText(str(len(self.gui.model.sampling_volume.get_points())))
        else:
            self.total_label.setText("N/A")



""" BiotSavart module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
from functools import partial
from PyQt5.QtCore import QThread
from magneticalc.Debug import Debug
from magneticalc.Metric import Metric


class BiotSavart:
    """
    Implements the Biot-Savart law for calculating the magnetic flux density.
    Employing static & class methods only, for better multiprocessing performance.
    """

    # Constant for Biot-Savart law, in SI units
    k = 1e-7  # H / m

    # Divisor cutoff (avoiding divisions by zero)
    DivisorCutoff = 1e-12

    # Class attributes
    dc = None
    current_elements = None
    sampling_volume_points = None
    progress_callback = None

    @classmethod
    def init(cls, dc, current_elements, sampling_volume_points, progress_callback):
        """
        Populates class attributes.

        @param dc: DC factor
        @param current_elements: List of current elements (list of 3D vector pairs: (element center, element direction))
        @param sampling_volume_points: List of sampling volume points
        @param progress_callback: Progress callback
        """
        cls.dc = dc
        cls.current_elements = current_elements
        cls.sampling_volume_points = sampling_volume_points
        cls.progress_callback = progress_callback

    @staticmethod
    def worker(current_elements, sampling_volume_point):
        """
        Calculates the magnetic flux density at some sampling volume point using the Biot-Savart law.

        @param current_elements: List of current elements (list of 3D vector pairs: (element center, element direction))
        @param sampling_volume_point: Sampling volume point (3D vector)
        @return: Magnetic flux density vector (3D vector)
        """
        vector = np.zeros(3)

        for element_center, element_direction in current_elements:
            vector_distance = (sampling_volume_point - element_center) * Metric.LengthScale
            scalar_distance_cubed = np.linalg.norm(vector_distance) ** 3

            # Avoiding divisions by zero
            if scalar_distance_cubed < BiotSavart.DivisorCutoff:
                continue

            vector += np.cross(element_direction, vector_distance) / scalar_distance_cubed

        return vector

    @classmethod
    def get_vectors(cls, pool):
        """
        Calculates the magnetic flux density at every point of the sampling volume.

        @return: List of vectors if successful, None if interrupted
        """
        Debug(cls, ".get_vectors()", color=(0, 0, 255))

        # Map sampling volume points to worker function, passing current elements as constant argument
        result = pool.imap(partial(cls.worker, cls.current_elements), cls.sampling_volume_points)

        vectors = []

        # Fetch resulting vectors
        for i, point in enumerate(result):

            vectors.append(point)

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                cls.progress_callback(100 * (i + 1) / len(cls.sampling_volume_points))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(cls, ": Interruption requested, exiting now", color=(0, 0, 255))
                    return None

        # Apply Biot-Savart constant scaling and DC factor
        vectors = np.array(vectors) * BiotSavart.k * BiotSavart.dc

        return vectors



""" Display_Widget module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QCheckBox
from magneticalc.Groupbox import Groupbox
from magneticalc.HLine import HLine
from magneticalc.IconLabel import IconLabel
from magneticalc.SliderFloat import SliderFloat


class Display_Widget(Groupbox):
    """ Display_Widget class. """

    # SliderFloat limits
    FieldArrowScaleMinimum = 0
    FieldArrowScaleMaximum = 1
    FieldArrowScaleStep = .1
    FieldPointScaleMinimum = 0
    FieldPointScaleMaximum = 10
    FieldPointScaleStep = 1
    FieldBoostMinimum = 0
    FieldBoostMaximum = 1
    FieldBoostStep = 0.05

    def __init__(self, gui):
        """
        Populates the widget.

        @param gui: GUI
        """
        Groupbox.__init__(self, "Display")

        self.gui = gui

        self.addWidget(IconLabel("fa.circle", "Point Scale"))
        self.field_point_scale_slider = SliderFloat(Qt.Horizontal)
        self.field_point_scale_slider.set_range_step(
            self.FieldPointScaleMinimum,
            self.FieldPointScaleMaximum,
            self.FieldPointScaleStep
        )
        self.field_point_scale_slider.setValue(self.gui.config.get_float("field_point_scale"))
        self.field_point_scale_slider.valueChanged.connect(
            lambda: self.set_field_point_scale(self.field_point_scale_slider.get_value())
        )
        self.addWidget(self.field_point_scale_slider)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("fa.arrow-right", "Arrow Scale"))
        self.field_arrow_scale_slider = SliderFloat(Qt.Horizontal)
        self.field_arrow_scale_slider.set_range_step(
            self.FieldArrowScaleMinimum,
            self.FieldArrowScaleMaximum,
            self.FieldArrowScaleStep
        )
        self.field_arrow_scale_slider.setValue(self.gui.config.get_float("field_arrow_scale"))
        self.field_arrow_scale_slider.valueChanged.connect(
            lambda: self.set_field_arrow_scale(self.field_arrow_scale_slider.get_value())
        )
        self.addWidget(self.field_arrow_scale_slider)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("fa.adjust", "Field Boost"))
        self.field_boost_slider = SliderFloat(Qt.Horizontal)
        self.field_boost_slider.set_range_step(
            self.FieldBoostMinimum,
            self.FieldBoostMaximum,
            self.FieldBoostStep
        )
        self.field_boost_slider.setValue(self.gui.config.get_float("field_boost"))
        self.field_boost_slider.valueChanged.connect(
            lambda: self.set_field_boost(self.field_boost_slider.get_value())
        )
        self.addWidget(self.field_boost_slider)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(HLine())

        self.addWidget(IconLabel("fa.tags", "Field Labels"))
        self.display_magnitude_labels_checkbox = QCheckBox(" Display Magnitude")  # Leading space for alignment
        self.display_magnitude_labels_checkbox.setChecked(self.gui.config.get_bool("display_magnitude_labels"))
        self.display_magnitude_labels_checkbox.stateChanged.connect(
            lambda: self.set_display_magnitude_labels(self.display_magnitude_labels_checkbox.isChecked())
        )
        self.addWidget(self.display_magnitude_labels_checkbox)

    def set_field_arrow_scale(self, value: float):
        """
        Sets field arrow scale.

        @param value: Value
        """
        self.gui.config.set_float("field_arrow_scale", value)
        self.gui.redraw()

    def set_field_point_scale(self, value: float):
        """
        Sets field point scale.

        @param value: Value
        """
        self.gui.config.set_float("field_point_scale", value)
        self.gui.redraw()

    def set_field_boost(self, value: float):
        """
        Sets field boost value.

        @param value: Value
        """
        self.gui.config.set_float("field_boost", value)
        self.gui.redraw()

    def set_display_magnitude_labels(self, value: bool):
        """
        Sets "Display Magnitude" value.

        @param value: Value
        """
        self.gui.config.set_bool("display_magnitude_labels", value)
        self.gui.redraw()



""" Table module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from functools import partial
import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QHeaderView, QPushButton, QAbstractItemView


class Table(QTableWidget):
    """ Table class. """

    def __init__(self, enabled=True, cell_edited_callback=None, row_deleted_callback=None, minimum_rows=2):
        """
        Initializes table.

        @param cell_edited_callback: Set this to make cells editable
        @param row_deleted_callback: Set this to make rows deletable
        @param minimum_rows: Minimum number of rows (no further rows can be deleted)
        """
        QTableWidget.__init__(self)

        self.enabled = enabled
        self.cell_edited_callback = cell_edited_callback
        self.row_deleted_callback = row_deleted_callback
        self.minimum_rows = minimum_rows

        if self.cell_edited_callback is not None:
            self.itemChanged.connect(self.cell_edited_callback)

        self.setStyleSheet("""
            QTableCornerButton::section {
                border-width: 1px;
                border-style: solid;
                border-color: white silver silver white;
            }

            QTableWidget::item {
                padding: 2;
            }
        """)

        self.setAlternatingRowColors(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.selectionModel().selectionChanged.connect(self.selection_changed)

    def selection_changed(self, a, b):
        print(a, b)

    def set_horizontal_header(self, header):
        """
        Sets horizontal header.

        @param header: List of row header strings
        """
        if self.row_deleted_callback is not None:
            header.append("")

        self.setColumnCount(len(header))

        for col_index, label in enumerate(header):
            item = QTableWidgetItem(label)
            item.setTextAlignment(Qt.AlignCenter | Qt.AlignVCenter)
            self.setHorizontalHeaderItem(col_index, item)
            self.horizontalHeader().setSectionResizeMode(col_index, QHeaderView.Stretch)

    def set_vertical_header(self, header):
        """
        Sets vertical header.

        @param header: List of column header strings
        """
        self.setRowCount(len(header))

        for row_index, label in enumerate(header):
            item = QTableWidgetItem(label)
            item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.setVerticalHeaderItem(row_index, item)
            self.verticalHeader().setSectionResizeMode(row_index, QHeaderView.ResizeToContents)

        self.verticalHeader().setStyleSheet("""
            QHeaderView::section {
                padding: 0 3px;
            }
        """)

    def clear_rows(self):
        """
        Clears table rows.
        """
        self.setRowCount(0)

    def set_contents(self, contents):
        """
        Sets table contents.

        @param contents: 2D array
        """
        self.blockSignals(True)

        for row_index, row_contents in enumerate(contents):
            for col_index, col_contents in enumerate(row_contents):

                item = QTableWidgetItem(col_contents)
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)

                flags = Qt.NoItemFlags | Qt.ItemIsSelectable
                if self.enabled:
                    flags |= Qt.ItemIsEnabled
                if self.cell_edited_callback is not None:
                    flags |= Qt.ItemIsEditable
                item.setFlags(flags)

                self.setItem(row_index, col_index, item)

            if self.row_deleted_callback is not None:
                delete_button = QPushButton()
                delete_button.setIcon(qta.icon("fa.minus"))
                delete_button.setStyleSheet("""
                    border: none;
                    background: palette(window);
                """)

                if self.rowCount() <= self.minimum_rows:
                    delete_button.setEnabled(False)
                else:
                    delete_button.clicked.connect(partial(self.row_deleted_callback, row_index))

                self.setCellWidget(row_index, self.columnCount() - 1, delete_button)

        self.blockSignals(False)



""" Config module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import configparser
from magneticalc.Debug import Debug
from magneticalc.VispyCanvas import VispyCanvas
from magneticalc.Wire import Wire


class Config:
    """ Config class. """

    # Configuration filename
    Filename = "MagnetiCalc.ini"

    # Enable to instantly save every change to file (useful for debugging)
    InstantSave = False

    # Enable to additionally debug read access to configuration
    DebugGetters = False

    # Default wire preset
    DefaultWirePreset = "Straight Line"

    # Default perspective preset
    DefaultPerspectivePreset = "Isometric"

    # Default configuration
    Default = {
        "window_width"                  : "1500",
        "window_height"                 : "1000",
        "auto_calculation"              : "True",
        "num_cores"                     : "0",
        "wire_points_base"              : None,  # Will be set in __init__
        "wire_stretch"                  : "0.3, 1.0, 1.0",
        "wire_slicer_limit"             : "0.1",
        "wire_dc"                       : "1.0",
        "rotational_symmetry_count"     : "10",
        "rotational_symmetry_radius"    : "1.0",
        "rotational_symmetry_axis"      : "2",
        "sampling_volume_padding"       : "1, -2, 2",
        "sampling_volume_resolution"    : "10",
        "color_metric"                  : "Magnitude",
        "alpha_metric"                  : "Log Magnitude",
        "field_point_scale"             : "1.0",
        "field_arrow_scale"             : "0.2",
        "field_boost"                   : "0.1",
        "display_magnitude_labels"         : "True",
        "show_wire_segments"            : "True",
        "show_wire_points"              : "True",
        "field_colors_labels"           : "False",
        "show_coordinate_system"        : "True",
        "show_perspective_info"         : "True",
        "dark_background"               : "True",
        "azimuth"                       : None,  # Will be set in __init__
        "elevation"                     : None,  # Will be set in __init__
        "scale_factor"                  : "5.0"
    }

    def __init__(self):
        """
        Loads configuration from file.
        """
        self.absolute_filename = os.path.join(os.path.dirname(os.path.dirname(__file__)), self.Filename)

        Debug(self, ": file://" + self.absolute_filename.replace(" ", "%20"), force=True)

        self.Default["wire_points_base"] = Config.points_to_str(Wire.get_by_id(Config.DefaultWirePreset)["points"])
        self.Default["azimuth"] = VispyCanvas.get_by_id(Config.DefaultPerspectivePreset)["azimuth"]
        self.Default["elevation"] = VispyCanvas.get_by_id(Config.DefaultPerspectivePreset)["elevation"]

        self.config = configparser.ConfigParser()
        self.synced = False

        self.load()

    def close(self):
        """
        Finally auto-saves configuration to file.
        """
        Debug(self, ".close()")

        if not self.synced:
            self.save()

    def load(self):
        """
        Loads the configuration from file.
        """
        Debug(self, ".load: " + self.absolute_filename)

        self.config.read(self.absolute_filename)
        self.set_defaults()
        self.synced = True

    def set_defaults(self):
        """
        Sets the default key-value pairs. Creates empty "User" section if not present.
        """
        self.config["DEFAULT"] = Config.Default

        if "User" not in self.config:
            Debug(self, ".set_defaults(): Creating empty User section")
            self.config["User"] = {}

    def save(self):
        """
        Saves the configuration to file.
        """
        Debug(self, ".save()")

        with open(self.Filename, "w") as file:
            self.config.write(file)
        self.synced = True

    def get_bool(self, key):
        """
        Gets boolean value, convert from string.

        @param key: Key
        @return: Value
        """
        return self.get_str(key) == "True"

    def get_float(self, key):
        """
        Gets float value, convert from string.

        @param key: Key
        @return: Value
        """
        return float(self.get_str(key))

    def get_int(self, key):
        """
        Gets integer value, convert from string.

        @param key: Key
        @return: Value
        """
        return int(self.get_str(key))

    def get_points(self, key):
        """
        Gets list of 3D points, convert from string.

        @param key: Key
        @return: List of 3D points
        """
        return Config.str_to_points(self.get_str(key))

    def get_point(self, key):
        """
        Gets a single 3D point, convert from string.

        @param key: Key
        @return: Single 3D point
        """
        return Config.str_to_point(self.get_str(key))

    def get_str(self, key):
        """
        Reads a configuration value. Key must be in "Default" section and may be overridden in "User" section.

        @param key: Key (string)
        @return: Value (string)
        """
        if self.DebugGetters:
            Debug(self, f".get_str({key})")

        value = self.config.get("User", key)
        return value

    def set_bool(self, key, value):
        """
        Sets boolean value, convert to string.

        @param key: Key
        @param value: Value
        """
        self.set_str(key, "True" if value else "False")

    def set_float(self, key, value):
        """
        Sets float value, convert to string.

        @param key: Key
        @param value: Value
        """
        self.set_str(key, f"{float(value):.3f}")

    def set_int(self, key, value):
        """
        Sets integer value, convert to string.

        @param key: Key
        @param value: Value
        """
        self.set_str(key, str(int(value)))

    def set_points(self, key, value):
        """
        Sets list of 3D points, convert to string.

        @param key: Key
        @param value: List of 3D points
        """
        self.set_str(key, Config.points_to_str(value))

    def set_point(self, key, value):
        """
        Sets a single 3D point, convert to string.

        @param key: Key
        @param value: Single 3D point
        """
        self.set_str(key, Config.point_to_str(value))

    def set_str(self, key, value):
        """
        Writes a configuration value. Key must be in "Default" section and may be overridden in "User" section.

        @param key: Key (string)
        @param value: Value (string)
        """
        Debug(self, f".set_str({key}, {value})")

        self.config.set("User", key, value)
        self.synced = False

        if self.InstantSave:
            self.save()

    def set_get_bool(self, key, value):
        """
        If value is not None, writes and returns key-value; Otherwise (if value is None), reads and returns key-value.

        @param key: Key (string)
        @param value: Value (bool) or None
        """
        if value is None:
            return self.get_bool(key)
        else:
            self.set_bool(key, value)
            return value

    def set_get_float(self, key, value):
        """
        If value is not None, writes and returns key-value; Otherwise (if value is None), reads and returns key-value.

        @param key: Key (string)
        @param value: Value (float) or None
        """
        if value is None:
            return self.get_float(key)
        else:
            self.set_float(key, value)
            return value

    def set_get_int(self, key, value):
        """
        If value is not None, writes and returns key-value; Otherwise (if value is None), reads and returns key-value.

        @param key: Key (string)
        @param value: Value (int) or None
        """
        if value is None:
            return self.get_int(key)
        else:
            self.set_int(key, value)
            return value

    def set_get_points(self, key, value):
        """
        If value is not None, writes and returns key-value; Otherwise (if value is None), reads and returns key-value.

        @param key: Key (string)
        @param value: Value (string) or None
        """
        if value is None:
            return self.get_points(key)
        else:
            self.set_points(key, value)
            return value

    def set_get_point(self, key, value):
        """
        If value is not None, writes and returns key-value; Otherwise (if value is None), reads and returns key-value.

        @param key: Key (string)
        @param value: Value (string) or None
        """
        if value is None:
            return self.get_point(key)
        else:
            self.set_point(key, value)
            return value

    def set_get_str(self, key, value):
        """
        If value is not None, writes and returns key-value; Otherwise (if value is None), reads and returns key-value.

        @param key: Key (string)
        @param value: Value (string) or None
        """
        if value is None:
            return self.get_str(key)
        else:
            self.set_str(key, value)
            return value

    @staticmethod
    def point_to_str(point):
        """
        Converts a single 3D point to string.

        @param point: Single 3D point
        @return: String
        """
        return f"{point[0]:+0.03f}, {point[1]:+0.03f}, {point[2]:+0.03f}"

    @staticmethod
    def points_to_str(points):
        """
        Converts list of 3D points to string.

        @param points: List of 3D points
        @return: String
        """
        return "; ".join([Config.point_to_str(point) for point in points])

    @staticmethod
    def str_to_point(str_point):
        """
        Converts string to single 3D point.

        @param str_point: String
        @return: Single 3D point
        """
        return [float(coord) for coord in str_point.split(",")]

    @staticmethod
    def str_to_points(str_points):
        """
        Converts string to list of 3D points.

        @param str_points: String
        @return: List of 3D points
        """
        return [Config.str_to_point(str_point) for str_point in str_points.split(";")]



""" Field module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from multiprocessing import Pool
from magneticalc.BiotSavart import BiotSavart
from magneticalc.Debug import Debug


class Field:
    """ Field class. """

    def __init__(self):
        """
        Initializes an empty field.
        """
        Debug(self, ": Init")

        self._vectors = None

    def is_valid(self):
        """
        Indicates valid data for display.

        @return: True if data is valid for display, False otherwise
        """
        return self._vectors is not None

    def invalidate(self):
        """
        Resets data, hiding from display.
        """
        Debug(self, ".invalidate()", color=(128, 0, 0))

        self._vectors = None

    def get_vectors(self):
        """
        Get field vectors.

        @return: List of 3D vectors
        """
        return self._vectors

    def recalculate(self, wire, sampling_volume, progress_callback, num_cores):
        """
        Recalculate field vectors.

        @param wire: Wire
        @param sampling_volume: Sampling volume
        @param progress_callback: Progress callback
        @param num_cores: Number of cores to use for multiprocessing
        @return: True if successful, False if interrupted
        """
        biot_savart = BiotSavart()

        # Initialize Biot-Savart class
        biot_savart.init(
            wire.get_dc(),
            wire.get_elements(),
            sampling_volume.get_points(),
            progress_callback
        )

        # Fetch multiprocessing result
        with Pool(num_cores) as pool:
            vectors = biot_savart.get_vectors(pool)

        # Handle interrupt
        if vectors is None:
            return False

        self._vectors = vectors

        return True



""" About_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import webbrowser
import qtawesome as qta
from functools import partial
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QTextBrowser, QPushButton
from magneticalc.Version import Version


class About_Dialog(QDialog):
    """ About_Dialog class. """

    # Window dimensions
    Width = 530
    Height = 375

    # HTML content
    HTML = f"""
        <span style="color: #2a7db0;"><b>{Version.String}</b></span><br>
        <br>
        Copyright © 2020, Paul Wilhelm, M. Sc.
        &lt;<a href="mailto:anfrage@paulwilhelm.de">anfrage@paulwilhelm.de</a>&gt;<br>
        <br><br>
        <small>
            <b>ISC License</b><br>
            <br>
            Permission to use, copy, modify, and/or distribute this software for any<br>
            purpose with or without fee is hereby granted, provided that the above<br>
            copyright notice and this permission notice appear in all copies.<br>
            <br>
            THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES<br>
            WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF<br>
            MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR<br>
            ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES<br>
            WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN<br>
            ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF<br>
            OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.<br>
        </small>
        <br><br>
        <span style="color: #2a7db0;">
            If you like this software, please consider buying me a coffee!&nbsp; ;)
        </span>
        """

    # Donation URL
    DonateURL = "https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=TN6YTPVX36YHA&source=url"

    def __init__(self):
        """
        Displays "About" dialog.
        """

        # noinspection PyArgumentList
        QDialog.__init__(self)

        self.setWindowTitle("About")

        layout = QVBoxLayout()
        self.setLayout(layout)

        text_browser = QTextBrowser()
        text_browser.setMinimumWidth(self.Width)
        text_browser.setMinimumHeight(self.Height)
        text_browser.setStyleSheet("""
            background: palette(window);
            border: none;
            line-height: 20px;
        """)
        text_browser.setOpenExternalLinks(True)
        text_browser.insertHtml(About_Dialog.HTML)
        text_browser.setFocusPolicy(Qt.NoFocus)
        layout.addWidget(text_browser)

        button_box = QHBoxLayout()

        ok_button = QPushButton(qta.icon("fa.check"), "OK")
        ok_button.clicked.connect(self.accept)
        button_box.addWidget(ok_button)

        donate_button = QPushButton(qta.icon("fa.paypal"), "Donate 3€")
        donate_button.clicked.connect(partial(webbrowser.open, About_Dialog.DonateURL))
        button_box.addWidget(donate_button)

        layout.addLayout(button_box)

    def show(self):
        """
        Shows this dialog.
        """
        self.exec()



""" MagnetiCalc package initialization. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# "Intentionally left blank"



""" Groupbox module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPalette, QColor
from PyQt5.QtWidgets import QGroupBox, QVBoxLayout


class Groupbox(QGroupBox):
    """ Groupbox class. """

    def __init__(self, title):
        """
        Initializes the groupbox.

        @param title: Title
        """
        QGroupBox.__init__(self)

        self.setTitle(title)
        self.setStyleSheet("""
            QGroupBox {
                border: 1px solid #cccccc;
                border-radius: 3px;
                margin-top: 20px;
                color: #2a7db0;
                font-weight: bold;
                background-color: #e5e5e5;
            }

            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                background-color: palette(window);
            }
        """)

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.palette = self.palette()
        self.palette.setColor(QPalette.Button, QColor(3, 18, 14))
        self.setPalette(self.palette)

    # noinspection PyPep8Naming
    def addWidget(self, widget):
        """
        Adds widget to groupbox.

        @param widget: QWidget
        """
        self.layout.addWidget(widget)

    # noinspection PyPep8Naming
    def addSpacing(self, spacing):
        """
        Adds spacing to groupbox.

        @param spacing: Spacing value
        """
        self.layout.addSpacing(spacing)

    # noinspection PyPep8Naming
    def addLayout(self, layout):
        """
        Adds layout to groupbox.

        @param layout: QLayout
        """
        self.layout.addLayout(layout)



""" IconLabel module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import qtawesome as qta
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QLabel


class IconLabel(QWidget):
    """ IconLabel class. """

    IconSize = QSize(16, 16)
    HorizontalSpacing = 1

    def __init__(self, qta_id, text):
        """
        Initializes the icon label.

        @param qta_id: QtAwesome icon id
        @param text: Label text
        """
        # noinspection PyArgumentList
        QWidget.__init__(self)

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        if qta_id is not None:
            icon = QLabel()
            icon.setPixmap(qta.icon(qta_id).pixmap(self.IconSize))
            self.layout.addWidget(icon, alignment=Qt.AlignVCenter)
            self.layout.addSpacing(self.HorizontalSpacing)

        label = QLabel(text)
        label.setStyleSheet("""
            font-weight: bold;
        """)
        self.layout.addWidget(label, alignment=Qt.AlignVCenter)

        self.layout.addStretch()

    # noinspection PyPep8Naming
    def addWidget(self, widget):
        """
        Adds widget to the horizontal layout

        @param widget: QWidget
        """
        self.layout.addWidget(widget, alignment=Qt.AlignVCenter)

    # noinspection PyPep8Naming
    def addStretch(self):
        """
        Adds stretch to the horizontal layout
        """
        self.layout.addStretch()

    # noinspection PyPep8Naming
    def addSpacing(self, value):
        """
        Adds spacing to the horizontal layout

        @param value: Spacing value
        """
        self.layout.addSpacing(value)



""" Sampling volume module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
from PyQt5.QtCore import QThread
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Debug import Debug


class SamplingVolume:
    """ Sampling volume class. """

    def __init__(self, resolution: int):
        """
        Initializes an empty sampling volume, with zero bounds and no constraints.

        @param resolution: Resolution
        """
        Debug(self, ": Init")

        self._resolution = resolution

        self.constraints = []

        self._bounds_min = np.zeros(3)
        self._bounds_max = np.zeros(3)

        self._points = None

        Assert_Dialog(resolution > 0, "Resolution must be > 0")

    def is_valid(self):
        """
        Indicates valid data for display.

        @return: True if data is valid for display, False otherwise
        """
        return self._points is not None

    def invalidate(self):
        """
        Resets data, hiding from display.
        """
        Debug(self, ".invalidate()", color=(128, 0, 0))

        self._points = None

    # ------------------------------------------------------------------------------------------------------------------

    def get_resolution(self):
        """
        Returns this volume's resolution.

        @return: Resolution
        """
        return self._resolution

    def get_bounds(self):
        """
        Returns this volume's bounding box.

        @return: _bounds_min, _bounds_max
        """
        return self._bounds_min, self._bounds_max

    def get_points(self):
        """
        Returns this sampling volume's points.

        @return: List of 3D points
        """
        Assert_Dialog(self.is_valid(), "Accessing invalidated sampling volume")

        return self._points

    # ------------------------------------------------------------------------------------------------------------------

    def set_bounds_nearest(self, bounds_min, bounds_max):
        """
        Adjusts this volume's bounding box to fully enclose a 3D wire curve.
        This expands the bounding box to the next integer grid coordinates.

        Note: This will not automatically invalidate the sampling volume

        @param bounds_min: Minimum bounding box point
        @param bounds_max: Maximum bounding box point
        @return: Rounded (_bounds_min, _bounds_max)
        """
        self._bounds_min = np.array([np.floor(x) for x in bounds_min])
        self._bounds_max = np.array([np.ceil(x) for x in bounds_max])

    def set_padding(self, dx, dy, dz):
        """
        Shrinks or enlarges this volume's bounding box by some amount, in each direction, symmetrically.

        Note: This will not automatically invalidate the sampling volume

        @param dx: Amount of padding in X-direction.
        @param dy: Amount of padding in Y-direction.
        @param dz: Amount of padding in Z-direction.
        """
        self._bounds_min -= np.array([dx, dy, dz])
        self._bounds_max += np.array([dx, dy, dz])

    # ------------------------------------------------------------------------------------------------------------------

    def add_constraint(self, lambda_expression):
        """
        Adds some constraint to this volume's point generator.

        @param lambda_expression: Lambda expression, evaluating to boolean (dict as argument, see L{constraint_dict})
        """
        Debug(self, f".add_constraint()")

        self.constraints.append(lambda_expression)

    @staticmethod
    def constraint_dict(x, y, z):
        """
        Returns a parameters for a given point, containing its coordinates, 3D radius and 2D radii in different planes.

        @param x: X-coordinate
        @param y: Y-coordinate
        @param z: Z-coordinate
        @return: Constraint variables parameters
        """
        return {
            "x": x,
            "y": y,
            "z": z,
            "radius": np.linalg.norm(np.array([x, y, z])),
            "radius_xy": np.linalg.norm(np.array([x, y])),
            "radius_xz": np.linalg.norm(np.array([x, z])),
            "radius_yz": np.linalg.norm(np.array([y, z]))
        }

    # ------------------------------------------------------------------------------------------------------------------

    def recalculate(self, progress_callback):
        """
        Recalculates the sampling volume points.

        @return: True if successful, False if interrupted
        """
        Debug(self, ".recalculate()", color=(0, 128, 0))

        # Calculate all possible grid coordinates
        coordinates = [[], [], []]
        for i in range(3):
            steps = np.ceil((self._bounds_max[i] - self._bounds_min[i]) * self._resolution).astype(int) + 1
            coordinates[i] = np.linspace(self._bounds_min[i], self._bounds_max[i], steps)

        # Calculate the total number of possible grid points

        points = []

        # Iterate through all possible grid points
        # Note: This loop maps the linearized "i" array index onto the cuboid "x, y, z" grid index
        span = np.array([len(axis) for axis in coordinates])
        n = span[0] * span[1] * span[2]
        x, y, z = 0, 0, 0
        for i in range(n):

            point = np.array([coordinates[0][x], coordinates[1][y], coordinates[2][z]])

            # Calculate inclusion relation
            inclusion = True
            for constraint in self.constraints:
                if not constraint(SamplingVolume.constraint_dict(*point)):
                    inclusion = False
                    break
            if inclusion:
                points.append(point)

            # Move to the next "x, y, z" grid index
            if x + 1 < span[0]:
                x += 1
            else:
                x = 0
                if y + 1 < span[1]:
                    y += 1
                else:
                    y = 0
                    z += 1

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                progress_callback(100 * (i + 1) / n)

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): Interruption requested, exiting now", color=(0, 0, 255))
                    return False

        self._points = np.array(points)

        Debug(
            self,
            ".recalculate(): "
            f"Calculated {len(self.constraints)} constraints "
            f"over {n} possible points, "
            f"yielded {len(self._points)} points",
            color=(0, 0, 255)
        )

        return True



""" CalculationThread module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtCore import QThread, pyqtSignal
from multiprocessing import cpu_count


class CalculationThread(QThread):
    """
    CalculationThread class.
    The calculation thread implements an offloaded pipeline of subsequent calculation subtasks.
    The pipeline may be interrupted at any time, thereby exiting this thread.
    """

    # Progress update signal
    progress_update = pyqtSignal(int)

    def __init__(self, gui):
        """
        Initializes calculation thread.

        @param gui: GUI
        """
        QThread.__init__(self)

        self.gui = gui

        # Connect progress update signal and create callback
        self.progress_update.connect(lambda x: self.gui.statusbar.progressbar.setValue(x))
        self.progress_callback = self.progress_update.emit

    def run(self):
        """
        Thread main function.
        """

        if not self.gui.model.wire.is_valid():
            self.gui.calculation_status.emit("Calculating Wire Segments...")

            if not self.gui.model.calculate_wire(self.progress_callback):
                self.trigger_finished(False)
                return

            self.trigger_on_wire_valid()

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        if not self.gui.model.sampling_volume.is_valid():
            self.gui.calculation_status.emit("Calculating Sampling Volume...")

            if not self.gui.model.calculate_sampling_volume(self.progress_callback):
                self.trigger_finished(False)
                return

            self.trigger_on_sampling_volume_valid()

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        if not self.gui.model.field.is_valid():
            self.gui.calculation_status.emit("Calculating Magnetic Flux Density...")

            num_cores = self.gui.config.get_int("num_cores")
            if num_cores == 0:
                # "Auto" setting
                num_cores = max(1, cpu_count() - 1)

            success = self.gui.model.calculate_field(self.progress_callback, num_cores)

            if not success:
                self.trigger_finished(False)
                return

            self.trigger_on_field_valid()

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        if not self.gui.model.metric.is_valid():
            self.gui.calculation_status.emit("Calculating Metric...")

            if not self.gui.model.calculate_metric(self.progress_callback):
                self.trigger_finished(False)
                return

            self.trigger_on_metric_valid()

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.trigger_finished(True)

    # ------------------------------------------------------------------------------------------------------------------

    def trigger_finished(self, success):
        """
        Signals that the calculation finished.

        @param success: True if calculation was successful, False otherwise
        """

        # We cannot directly call this; we won't be able to modify the UI thread (which we'd really like to do somehow)
        # self.gui.calculation_stopped(success)

        # Instead, we use a signal; however, this results in slightly delayed execution, even after joining this thread;
        # the execution of  self.gui.on_calculation_exited()  will later be skipped when it sees another thread running.
        self.gui.calculation_exited.emit(success)

    # ------------------------------------------------------------------------------------------------------------------

    def trigger_on_wire_valid(self):
        """
        Gets called when the wire was successfully calculated.
        """
        self.gui.sidebar_left.wire_widget.update_sliced_total_label()
        self.gui.vispy_canvas.redraw()

    def trigger_on_sampling_volume_valid(self):
        """
        Gets called when the sampling volume was successfully calculated.
        """
        self.gui.sidebar_left.sampling_volume_widget.update_total_label()
        self.gui.vispy_canvas.redraw()

    def trigger_on_field_valid(self):
        """
        Gets called when the field was successfully calculated.
        """
        self.gui.vispy_canvas.redraw()

    def trigger_on_metric_valid(self):
        """
        Gets called when the metric was successfully calculated.
        """
        self.gui.sidebar_right.metric_widget.update_labels()
        self.gui.vispy_canvas.redraw()



""" Usage_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import qtawesome as qta
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QTextBrowser, QPushButton


class Usage_Dialog(QDialog):
    """ Usage_Dialog class. """

    # Window dimensions
    Width = 835
    Height = 590

    # HTML content
    HTML = f"""
        <h3 style="color: #2a7db0;">First Steps</h3>
        <ul>
            <li>Go to <b>Load Wire Preset</b> to select a basic wire shape.</li>
            <li>Customize the wire's base points inside the spreadsheet.</li>
            <li>Apply some transformations to your basic wire; rotate or stretch its shape.</li>
            <li>Reduce the slicer limit to improve the B-field calculation accuracy.</li>
            <li>Set a DC current flowing through your wire.</li>
        </ul>

        <ul>
            <li>
                The sampling volume is the grid over which the B-field is calculated;<br>
                increase its resolution to the desired level to show every detail of the field.
            </li>
            <li>
                By default, the cuboid sampling volume covers the wire completely;<br>
                however, you may symmetrically adjust its bounding box by adding (subtracting) some padding.
            </li>
        </ul>

        <ul>
            <li>
                Select a color/alpha metric to adjust the field's hue/transparency individually.<br>
                Because all metrics are normalized, changes in DC current do not affect the field hue/transparency;<br>
                however, the displayed metric limits linearly scale with the DC current.
            </li>
        </ul>

        <ul>
            <li>Take a look at the minimum and maximum magnetic flux densities reached in each metric.</li>
        </ul>

        <ul>
            <li>Use the scroll wheel to zoom in and out of the 3D scene.</li>
            <li>Click and drag into the 3D scene to rotate.</li>
            <li>Press SHIFT while dragging to <i>move</i> the entire 3D scene.</li>
            <li>If you like the result, press CTRL+S to save a screenshot!
        </ul>

        All settings (including your wire shape) are stored in the <code>MagnetiCalc.ini</code> file.<br>
        Deleting or renaming this file will restore the default settings.

        <h3 style="color: #2a7db0;">What does MagnetiCalc do?</h3>

        MagnetiCalc calculates the magnetic field of arbitrary air coils, examples included.<br>
        Inside this VisPy/OpenGL-accelerated PyQt5 GUI, the static magnetic flux density (B-field due to a DC current)
        <br>
        is displayed in interactive 3D, using multiple metrics for highlighting this field's properties.<br>
        All parameters and presets can interactively be changed inside the GUI.

        <h3 style="color: #2a7db0;">Who needs MagnetiCalc?</h3>

        MagnetiCalc does its job for hobbyists, students, engineers and researchers of magnetic phenomena.<br>
        I designed MagnetiCalc from scratch, because I didn't want to mess around with expensive and/or<br>
        overly complex simulation software whenever I needed to solve a magnetostatic problem.

        <h3 style="color: #2a7db0;">How does it work?</h3>

        The field calculation is implemented using the Biot-Savart law <span style="color: #2a7db0;">[1]</span>,
        employing multiprocessing techniques.<br>
        The use of easily constrainable "sampling volumes" allows for selective calculation over arbitrary shapes.<br>
        <br>
        The shape of any wire is modeled as a 3D piecewise linear curve.<br>
        Arbitrary loops of wire are sliced into differential current elements,<br>
        each of which contributes to the total magnetic flux density B at some fixed 3D grid point.<br>
        <br>
        At each grid point, the field is displayed using colored arrows and/or dots;<br>
        field color and alpha transparency are individually mapped using one of the various available metrics.<br>
        <br>
        <span style="color: #2a7db0;">[1]</span>: Jackson, Klassische Elektrodynamik, 5. Auflage, S. 204, (5.4).<br>
        <br>
        <br>
        <span style="color: #2a7db0;">
            This and more information about MagnetiCalc can be found in the <code>README.md</code> file.
        </span>
        <br>
        """

    def __init__(self):
        """
        Displays "Usage" dialog.
        """

        # noinspection PyArgumentList
        QDialog.__init__(self)

        self.setWindowTitle("Usage")

        layout = QVBoxLayout()
        self.setLayout(layout)

        text_browser = QTextBrowser()
        text_browser.setMinimumWidth(self.Width)
        text_browser.setMinimumHeight(self.Height)
        text_browser.setStyleSheet("""
            background: palette(window);
            border: none;
            line-height: 20px;
        """)
        text_browser.setOpenExternalLinks(True)
        text_browser.insertHtml(Usage_Dialog.HTML)
        text_browser.setFocusPolicy(Qt.NoFocus)
        cursor = text_browser.textCursor()
        cursor.setPosition(0)
        text_browser.setTextCursor(cursor)
        layout.addWidget(text_browser)

        button_box = QHBoxLayout()
        ok_button = QPushButton(qta.icon("fa.check"), "OK")
        ok_button.clicked.connect(self.accept)
        button_box.addWidget(ok_button)
        layout.addLayout(button_box)

    def show(self):
        """
        Shows this dialog.
        """
        self.exec()



""" Wire module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
from PyQt5.QtCore import QThread
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Debug import Debug


class Wire:
    """ Wire class. """

    # Preset: A straight line.
    StraightLine = {
        "id": "Straight Line",
        "points": [
            [-1/2, 0, 0],
            [+1/2, 0, 0]
        ]
    }

    # Preset: A single loop with offset connections.
    SingleLoop_Offset = {
        "id": "Single Loop (offset)",
        "points": [
            [-1/2, +1/2, +1/2],
            [+0/2, +1/2, +1/2],
            [+0/2, -1/2, +1/2],
            [+0/2, -1/2, -1/2],
            [+0/2, +1/2, -1/2],
            [+0/2, +1/2, +1/2],
            [+1/2, +1/2, +1/2]
        ]
    }

    # Preset: A single loop with centered connections.
    SingleLoop_Centered = {
        "id": "Single Loop (centered)",
        "points": [
            [-1/2, +1/2, +0/2],
            [+0/2, +1/2, +0/2],
            [+0/2, +1/2, +1/2],
            [+0/2, -1/2, +1/2],
            [+0/2, -1/2, -1/2],
            [+0/2, +1/2, -1/2],
            [+0/2, +1/2, +0/2],
            [+1/2, +1/2, +0/2]
        ]
    }

    # Preset: A "compensated" double loop with offset connections.
    CompensatedDoubleLoop_Offset = {
        "id": "Compensated Double Loop (offset)",
        "points": [
            [-3/6, +1/2, +1/2],
            [-1/6, +1/2, +1/2],
            [-1/6, -1/2, +1/2],
            [-1/6, -1/2, -1/2],
            [-1/6, +1/2, -1/2],
            [-1/6, +1/2, +1/2],
            [+1/6, +1/2, +1/2],
            [+1/6, +1/2, -1/2],
            [+1/6, -1/2, -1/2],
            [+1/6, -1/2, +1/2],
            [+1/6, +1/2, +1/2],
            [+3/6, +1/2, +1/2]
        ]
    }

    # Preset: A "compensated" double loop with centered connections.
    CompensatedDoubleLoop_Centered = {
        "id": "Compensated Double Loop (centered)",
        "points": [
            [-3/6, +1/2, +0/2],
            [-1/6, +1/2, +0/2],
            [-1/6, +1/2, +1/2],
            [-1/6, -1/2, +1/2],
            [-1/6, -1/2, -1/2],
            [-1/6, +1/2, -1/2],
            [-1/6, +1/2, +0/2],
            [+1/6, +1/2, +0/2],
            [+1/6, +1/2, -1/2],
            [+1/6, -1/2, -1/2],
            [+1/6, -1/2, +1/2],
            [+1/6, +1/2, +1/2],
            [+1/6, +1/2, +0/2],
            [+3/6, +1/2, +0/2]
        ]
    }

    # List of all above presets
    Presets = [
        StraightLine,
        SingleLoop_Offset,
        SingleLoop_Centered,
        CompensatedDoubleLoop_Offset,
        CompensatedDoubleLoop_Centered
    ]

    @staticmethod
    def get_by_id(_id_):
        """
        Selects a preset by name.

        @param _id_: Preset ID string
        @return: Preset parameters (or None if ID not found)
        """
        for preset in Wire.Presets:
            if _id_ == preset["id"]:
                return preset
        return None

    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, points, stretch, rotational_symmetry, slicer_limit, dc):
        """
        A 3D piecewise linear curve with some DC current associated with it.

        @param points: List of 3D coordinates (see presets)
        @param stretch: XYZ stretch transform factors (3D point)
        @param rotational_symmetry: Dictionary for rotational symmetry transform
        @param slicer_limit: Slicer limit
        @param dc: DC current (float)
        """
        Debug(self, ": Init")

        self._points_base = np.array(points)

        self._slicer_limit = slicer_limit
        self._dc = dc

        self._points_sliced = None

        self._points_transformed = self._points_base.copy()

        self._set_stretch(stretch)
        self._set_rotational_symmetry(rotational_symmetry)

        Assert_Dialog(len(self._points_base) >= 2, "Number of points must be >= 2")

    def is_valid(self):
        """
        Indicates valid data for display.

        @return: True if data is valid for display, False otherwise
        """
        return self._points_sliced is not None

    def invalidate(self):
        """
        Resets data, hiding from display.
        """
        Debug(self, ".invalidate()", color=(128, 0, 0))

        self._points_sliced = None

    # ------------------------------------------------------------------------------------------------------------------

    def get_dc(self):
        """
        Returns DC value.

        @return: DC value
        """
        return self._dc

    def get_bounds(self):
        """
        Returns this curve's bounding box.

        @return: Minimum bounds (3D point), maximum bounds (3D point)
        """
        axes = self.get_points_transformed().transpose()
        bounds_min = [min(axes[0]), min(axes[1]), min(axes[2])]
        bounds_max = [max(axes[0]), max(axes[1]), max(axes[2])]
        return bounds_min, bounds_max

    def get_elements(self):
        """
        Returns this curve's elements, i.e. a list of segment center points and directions.

        @return: [(element_center, element_direction), ...]
        """
        Assert_Dialog(self.is_valid(), "Accessing invalidated wire")

        result = []
        for i in range(len(self._points_sliced) - 1):
            element_direction = np.array(self._points_sliced[i + 1]) - np.array(self._points_sliced[i])
            element_center = self._points_sliced[i] + element_direction / 2
            result.append((element_center, element_direction))
        return result

    def get_points_base(self):
        """
        Returns this wire's base points.

        @return: List of 3D points
        """
        return self._points_base

    def get_points_transformed(self):
        """
        Returns this wire's transformed points.

        @return: List of 3D points
        """
        return self._points_transformed

    def get_points_sliced(self):
        """
        Returns this wire's points after slicing.

        @return: List of 3D points
        """
        Assert_Dialog(self.is_valid(), "Accessing invalidated wire")

        return self._points_sliced

    # ------------------------------------------------------------------------------------------------------------------

    def _set_stretch(self, stretch):
        """
        This transformation stretches (and/or mirrors) this curve by some factor in any direction.
        Use the factor +1 / -1 to retain / mirror the curve in that direction.

        Note: Intended to be called from the class constructor (doesn't automatically invalidate the wire)

        @param stretch: XYZ stretch transform factors (3D point)
        """
        Debug(self, ".stretch()")

        print("A", self.get_points_base())
        print("B", self.get_points_transformed())

        axes = self.get_points_transformed().transpose()

        for i in range(3):
            axes[i] *= stretch[i]

        self._points_transformed = axes.transpose()

    def recalculate(self, progress_callback):
        """
        Slices wire segments into smaller ones until segment lengths equal or undershoot slicer limit.

        @param progress_callback: Progress callback
        @return: True if successful, False if interrupted
        """
        Debug(self, ".recalculate()", color=(0, 128, 0))

        points_sliced = []

        for i in range(len(self.get_points_transformed()) - 1):

            # Calculate direction and length of wire segment
            segment_direction = np.array(self.get_points_transformed()[i + 1] - self.get_points_transformed()[i])
            segment_length = np.linalg.norm(segment_direction)

            # Calculate required number of slices (subdivisions) and perform linear interpolation
            slices = np.ceil(segment_length / self._slicer_limit).astype(int)
            linear = np.linspace(0, 1, slices, endpoint=False)
            points_sliced += [self.get_points_transformed()[i] + segment_direction * j for j in linear]

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                progress_callback(100 * (i + 1) / (len(self.get_points_transformed()) - 1))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): Interruption requested, exiting now", color=(0, 0, 255))
                    return False

        # Close the loop
        points_sliced.append(self.get_points_transformed()[-1])

        self._points_sliced = np.array(points_sliced)

        return True

    # ------------------------------------------------------------------------------------------------------------------

    def _set_rotational_symmetry(self, parameters):
        """
        This transformation replicates and rotates this curve `count` times about an `axis` with radius `radius`.

        Note: Intended to be called from the class constructor (doesn't automatically invalidate the wire)

        @param parameters: Dictionary containing the transformation parameters (number of replications, radius and axis)
        """
        Debug(self, "._set_rotational_symmetry()")

        axes = self.get_points_transformed().transpose()

        x, y, z = [], [], []

        axis_other_1 = (parameters["axis"] + 1) % 3
        axis_other_2 = (parameters["axis"] + 2) % 3

        for a in np.linspace(0, 2 * np.pi, parameters["count"], endpoint=False):
            x = np.append(x, axes[axis_other_1] * np.sin(a) - (axes[axis_other_2] + parameters["radius"]) * np.cos(a))
            y = np.append(y, axes[axis_other_1] * np.cos(a) + (axes[axis_other_2] + parameters["radius"]) * np.sin(a))
            z = np.append(z, axes[parameters["axis"]])

        axes = [x, y, z]

        # Close the resulting loop
        for i in range(3):
            axes[i] = np.append(axes[i], axes[i][0])

        self._points_transformed = np.array(axes).transpose()

        return self



""" Model module. """

#  ISC License
#
#  Copyright (c) 2020, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.Debug import Debug
from magneticalc.Field import Field


class Model:
    """
    Model class.
    The model maintains a strict hierarchy of dependencies:  metric > field > sampling volume > wire
    When a lower module's data changed, all higher modules are invalidated (i.e. have their calculation results reset).
    """

    def __init__(self, on_metric_invalidated):
        """
        Initializes the model.

        @param on_metric_invalidated: Callback (needed to clear the metric labels after the metric became invalid)
        """
        Debug(self, ": Init")

        self.wire = None             # Will be initialized by Wire_Widget
        self.sampling_volume = None  # Will be initialized by SamplingVolume_Widget
        self.field = Field()         # There is no dedicated "Field_Widget"; set a single field here and reuse it
        self.metric = None
        self.on_metric_invalidated = on_metric_invalidated

    def is_valid(self):
        """
        Indicates valid data for display.

        @return: True if data is valid for display, False otherwise
        """
        return \
            self.wire.is_valid() and \
            self.sampling_volume.is_valid() and \
            self.field.is_valid() and \
            self.metric.is_valid()

    def invalidate(self, do_wire=False, do_sampling_volume=False, do_field=False, do_metric=False):
        """
        Invalidates multiple modules at once.

        @param do_wire: Enable to invalidate wire
        @param do_sampling_volume: Enable to invalidate sampling volume
        @param do_field: Enable to invalidate field
        @param do_metric: Enable to invalidate metric
        """
        if do_wire:
            if self.wire is not None:
                if self.wire.is_valid():
                    self.wire.invalidate()

        if do_sampling_volume:
            if self.sampling_volume is not None:
                if self.sampling_volume.is_valid():
                    self.sampling_volume.invalidate()

        if do_field:
            if self.field is not None:
                if self.field.is_valid():
                    self.field.invalidate()

        if do_metric:
            if self.metric is not None:
                if self.metric.is_valid():
                    self.metric.invalidate()
                    self.on_metric_invalidated()

    def set_wire(self, wire):
        """
        Sets the wire.

        @param wire: Wire
        """
        self.wire = wire
        self.invalidate(do_sampling_volume=True, do_field=True, do_metric=True)

    def set_sampling_volume(self, sampling_volume):
        """
        Sets the sampling volume.

        @param sampling_volume: Sampling volume
        """
        self.sampling_volume = sampling_volume
        self.invalidate(do_field=True, do_metric=True)

    def set_metric(self, metric):
        """
        Sets the metric.
        """
        self.metric = metric
        self.invalidate(do_metric=True)

    # ------------------------------------------------------------------------------------------------------------------

    def calculate_wire(self, progress_callback):
        """
        @return: True if successful, False if interrupted
        """
        Debug(self, ".calculate_wire()", color=(0, 0, 255))
        self.invalidate(do_sampling_volume=True, do_field=True, do_metric=True)
        return self.wire.recalculate(progress_callback)

    def calculate_sampling_volume(self, progress_callback):
        """
        @return: True if successful, False if interrupted
        """
        Debug(self, ".calculate_sampling_volume()", color=(0, 0, 255))
        self.invalidate(do_field=True, do_metric=True)
        return self.sampling_volume.recalculate(progress_callback)

    def calculate_field(self, progress_callback, num_cores):
        """
        @return: True if successful, False if interrupted
        """
        Debug(self, f".calculate_field(num_cores={num_cores})", color=(0, 0, 255))
        self.invalidate(do_metric=True)
        return self.field.recalculate(self.wire, self.sampling_volume, progress_callback, num_cores)

    def calculate_metric(self, progress_callback):
        """
        @return: True if successful, False if interrupted
        """
        Debug(self, ".calculate_metric()", color=(0, 0, 255))
        return self.metric.recalculate(self.field.get_vectors(), progress_callback)
